"""Public init."""

from .queue import Queue

__all__ = ["Queue"]

# NOTE: `__version__` is not defined because this package is built using 'setuptools-scm' --
#   use `importlib.metadata.version(...)` if you need to access version info at runtime.



"""Script for benchmarking different message compression options."""

import argparse
import json

TITLE_SPACING = 21
SIZE_SPACING = 8


def compare_size(name, compress, data):
    """Compare the size of the data in various compression methods."""

    was_bytes = False
    if isinstance(data, bytes):
        was_bytes = True
        print(f"{'raw bytes':{TITLE_SPACING}} {len(data):{SIZE_SPACING}}")
        import base64

        data = base64.b64encode(data).decode("utf-8")
        print("\nthe following use base64 first...\n")

    compressed_size = len(compress(json.dumps(data).encode()))
    print(f"{f'json+encode+{name}':{TITLE_SPACING}} {compressed_size:{SIZE_SPACING}}")

    json_encoded_size = len(json.dumps(data).encode())
    print(
        f"{'only json+encode':{TITLE_SPACING}} {json_encoded_size:{SIZE_SPACING}} (comp. ratio: {compressed_size / json_encoded_size:.3f})"
    )

    import pickle

    if not was_bytes:
        pickle_size = len(pickle.dumps(data))
        print(
            f"{'only pickled':{TITLE_SPACING}} {pickle_size:{SIZE_SPACING}} (comp. ratio: {compressed_size / pickle_size:.3f})"
        )


def compress_decompress(compress, decompress, data):
    if isinstance(data, bytes):
        import base64

        data = base64.b64encode(data).decode("utf-8")
    comp_data = compress(json.dumps(data).encode())
    out_data = json.loads(decompress(comp_data))
    # just in case there's a typo, check the data
    assert out_data == data


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--algo", required=True)
    parser.add_argument("--data", required=True)
    parser.add_argument("--only-size", action="store_true")
    args = parser.parse_args()

    # get data

    if args.data == "large_dict":
        data = {
            chr(i): [{chr(k % (i + 1)): k} for k in range(1000)] for i in range(100)
        }
    elif args.data == "medium_string":
        data = "\n".join(
            [
                "The quick brown fox jumps over the lazy dog.",
                "She sells sea shells by the sea shore.",
                "How much wood would a woodchuck chuck if a woodchuck could chuck wood?",
                "Peter Piper picked a peck of pickled peppers.",
                "A journey of a thousand miles begins with a single step.",
                "To be or not to be, that is the question.",
                "All that glitters is not gold.",
                "A picture is worth a thousand words.",
                "Actions speak louder than words.",
                "Better late than never.",
                "Birds of a feather flock together.",
                "Fortune favors the bold.",
                "A watched pot never boils.",
                "Beauty is in the eye of the beholder.",
                "Beggars can't be choosers.",
                "Cleanliness is next to godliness.",
                "Discretion is the better part of valor.",
                "Every cloud has a silver lining.",
                "Good things come to those who wait.",
                "Haste makes waste.",
                "If it ain't broke, don't fix it.",
                "Laughter is the best medicine.",
                "Necessity is the mother of invention.",
                "No man is an island.",
                "Practice makes perfect.",
                "Rome wasn't built in a day.",
                "The early bird catches the worm.",
                "The pen is mightier than the sword.",
            ]
        )
    elif args.data == "skyscan_i3_frame_pkl":
        data = b"\x80\x04\x95\x96\x02\x00\x00\x00\x00\x00\x00\x8c\x10icecube._icetray\x94\x8c\x07I3Frame\x94\x93\x94)R\x94}\x94Bh\x02\x00\x00[i3]\x06\x00\x00\x00\x00\x00P\x05\x00\x00\x00\r\x00\x00\x00I3EventHeader\r\x00\x00\x00I3EventHeader\x86\x00\x00\x00\x00\x01\x02\x00\r\x00\x00\x00I3EventHeader\x01\x03\x00\x00\x00\x00\x01\x00\x01\x00\x00\x00>\x16\x02\x00\x00\x00\x00\x00\x94\x88t\x00\x14\x00\x00\x00\x08\x00\x00\x00#\x00\x00\x00SCAN_nside0001_pixel0008_posvar0000\x01\x00\x02\x00\x00\x00\x03\x00\x00\x00\xe6\x07\x00\x00\xc1\x9d\xde^\xe8,\x18\x02\x04\x00\x00\x00\x05\x00\x00\x00\xe6\x07\x00\x00o\xa0\xe1^\xe8,\x18\x02\x15\x00\x00\x00MillipedeSeedParticle\n\x00\x00\x00I3Particle\x96\x00\x00\x00\x00\x01\x02\x00\n\x00\x00\x00I3Particle\x01\x05\x00\x00\x00\x00\x01\x00\x01\x00\x00\x00\x14\x00\x00\x00N\t\xb7q\xd0\xd0\x14U\x00\x00\x00\x00(\x00\x00\x00\x00\x00\x00\x00\x01\x00\x02\x00\x00\x00\x03\x00\x00\x00\xaeG\xe1z\x14\xdec@H\xe1z\x14\xae\x93u\xc0\n\xd7\xa3p=\x86\x7f\xc0\x01\x00\x04\x00\x00\x00\x05\x00\x00\x00\xa8aF\x06fj\x02@\x88\xe5:\xd1\xc7\x07\x0c@\x00\x00\x00\x00jr&A\x00\x00\x00\x00\x00\x00\xf8\x7f\x00\x00\x00\x00\x00\x00\xf8\x7f\x14\xa3\xac\xb4\xcc/\xd3?\x00\x00\x00\x00\x11\x00\x00\x00SCAN_HealpixNSide\x10\x00\x00\x00I3PODHolder<int>\x1d\x00\x00\x00\x00\x01\x02\x00\x05\x00\x00\x00I3Int\x01\x00\x00\x00\x00\x00\x01\x00\x01\x00\x00\x00\x01\x00\x00\x00\x11\x00\x00\x00SCAN_HealpixPixel\x10\x00\x00\x00I3PODHolder<int>\x1d\x00\x00\x00\x00\x01\x02\x00\x05\x00\x00\x00I3Int\x01\x00\x00\x00\x00\x00\x01\x00\x01\x00\x00\x00\x08\x00\x00\x00\x1b\x00\x00\x00SCAN_PositionVariationIndex\x10\x00\x00\x00I3PODHolder<int>\x1d\x00\x00\x00\x00\x01\x02\x00\x05\x00\x00\x00I3Int\x01\x00\x00\x00\x00\x00\x01\x00\x01\x00\x00\x00\x00\x00\x00\x00\x1aU\x07\xa1\x94\x86\x94b."
    else:
        raise ValueError("invalid --data value")

    # go!

    if args.algo == "bz2":
        import bz2

        if args.only_size:
            compare_size(args.algo, bz2.compress, data)
        else:
            compress_decompress(bz2.compress, bz2.decompress, data)

    elif args.algo == "lzma":
        import lzma

        if args.only_size:
            compare_size(args.algo, lzma.compress, data)
        else:
            compress_decompress(lzma.compress, lzma.decompress, data)

    elif args.algo == "zstd":
        import zstd  # type: ignore

        zstd_compress_1thread = lambda x: zstd.compress(  # noqa: E731
            x,
            3,  # level (3 is default)
            1,  # num of threads (auto is default)
        )
        if args.only_size:
            compare_size(args.algo, zstd_compress_1thread, data)
        else:
            compress_decompress(zstd_compress_1thread, zstd.decompress, data)

    elif args.algo == "gzip":
        import gzip

        if args.only_size:
            compare_size(args.algo, gzip.compress, data)
        else:
            compress_decompress(gzip.compress, gzip.decompress, data)

    elif args.algo == "lz4":
        import lz4.frame  # type: ignore

        if args.only_size:
            compare_size(args.algo, lz4.frame.compress, data)
        else:
            compress_decompress(lz4.frame.compress, lz4.frame.decompress, data)

    else:
        raise Exception("invalid compression algorithm specified")


if __name__ == "__main__":
    main()



"""Unit test imports."""

import importlib


def test_import() -> None:
    """Test module imports."""
    m = importlib.import_module("mqclient")
    print(dir(m))
    assert hasattr(m, "queue")
    assert hasattr(m, "broker_client_interface")



"""A server sends work out on one queue, and receives results on another."""

import argparse
import asyncio
import logging
import typing

import coloredlogs  # type: ignore[import]
from mqclient import Queue


async def server(work_queue: Queue, result_queue: Queue) -> None:
    """Demo example server."""
    async with work_queue.open_pub() as p:
        for i in range(100):
            await p.send({"id": i, "cmd": f'echo "{i}"'})

    results = {}
    result_queue.timeout = 5
    async with result_queue.open_sub() as stream:
        async for data in stream:
            assert isinstance(data, dict)
            results[typing.cast(int, data["id"])] = typing.cast(str, data["out"])

    print(results)
    assert len(results) == 100
    for i in results:
        assert results[i].strip() == str(i)


if __name__ == "__main__":
    coloredlogs.install(level=logging.DEBUG)

    parser = argparse.ArgumentParser(description="Worker")
    parser.add_argument("--broker-client", required=True, help="which kind of broker")
    parser.add_argument("--address", default="localhost", help="queue address")
    parser.add_argument("--work-queue", default="queue1", help="work queue")
    parser.add_argument("--result-queue", default="queue2", help="result queue")
    parser.add_argument(
        "--prefetch", type=int, default=10, help="result queue prefetch"
    )
    args = parser.parse_args()

    workq = Queue(
        args.broker_client,
        address=args.address,
        name=args.work_queue,
    )
    resultq = Queue(
        args.broker_client,
        address=args.address,
        name=args.result_queue,
        prefetch=args.prefetch,
    )

    asyncio.get_event_loop().run_until_complete(server(workq, resultq))



"""Define an interface that broker_clients will adhere to."""

import json
import uuid
from enum import Enum, auto
from typing import Any, AsyncGenerator, Dict, Optional, Union

import zstd  # type: ignore[import-not-found]  # false negative by mypy

from .config import MIN_PREFETCH

MessageID = Union[int, str, bytes]


class MQClientException(Exception):
    """Any exception for an error originating here."""


class ConnectingFailedException(MQClientException):
    """Raised when a `connect()` fails."""


class ClosingFailedException(MQClientException):
    """Raised when a `close()` fails."""


class AckException(MQClientException):
    """Raised when there's a problem with acking."""


class NackException(MQClientException):
    """Raised when there's a problem with nacking."""


class Message:
    """Message object.

    Holds msg_id and data.
    """

    class AckStatus(Enum):
        """Signify the ack state of a message."""

        NONE = auto()  # message has not been acked nor nacked
        ACKED = auto()  # message has been acked
        NACKED = auto()  # message has been nacked

    def __init__(self, msg_id: MessageID, payload: bytes):
        if not isinstance(msg_id, (int, str, bytes)):
            raise TypeError(
                f"Message.msg_id must be type int|str|bytes (not '{type(msg_id)}')."
            )
        if not isinstance(payload, bytes):
            raise TypeError(
                f"Message.data must be type 'bytes' (not '{type(payload)}')."
            )
        self.msg_id = msg_id
        self.payload = payload
        self._ack_status: Message.AckStatus = Message.AckStatus.NONE

        self._deserialized_payload: Dict[str, Any] = {}

        # set for special purposes since msg_id is not unique on redelivery
        self.uuid = int(uuid.uuid4())

        # set for special purposes which vary per broker api
        self._connection_id: Optional[int] = None

    def __repr__(self) -> str:
        """Return string of basic properties/attributes."""
        return f"Message(msg_id={self.msg_id!r}, payload={self.payload!r}, _ack_status={self._ack_status}, uuid={self.uuid}, _connection_id={self._connection_id})"

    def __eq__(self, other: object) -> bool:
        """Return True if self's and other's `data` are equal.

        On redelivery, `msg_id` may differ from its original, so
        `msg_id` is not a reliable source for testing equality. And
        neither is the `headers` field.
        """
        return bool(other) and isinstance(other, Message) and (self.data == other.data)

    @property
    def data(self) -> Any:
        """The object from the `data` field."""
        return self._deserialize()["data"]

    @property
    def headers(self) -> Any:
        """The dict from the `headers` field."""
        return self._deserialize()["headers"]

    def _deserialize(self) -> Dict[str, Any]:
        if not self._deserialized_payload:
            self._deserialized_payload = json.loads(zstd.decompress(self.payload))
        return self._deserialized_payload

    @staticmethod
    def serialize(data: Any, headers: Optional[Dict[str, Any]] = None) -> bytes:
        """Return serialized (bytes) representation of message payload.

        Optionally include `headers` dict for internal information.

        Data is compressed using `zstd.compress`, which was chosen by comparing
        the performance of bz2, lzma, zstd, gzip, and lz4 compression methods on
        various types of data.
        """
        if not headers:
            headers = {}

        return zstd.compress(
            json.dumps({"headers": headers, "data": data}).encode("utf-8"),
            3,  # level (3 is default)
            1,  # num of threads (auto is default)
        )


# -----------------------------
# classes to override/implement
# -----------------------------


class RawQueue:
    """Raw queue object, to hold queue state."""

    async def connect(self) -> None:
        """Set up connection."""

    async def close(self) -> None:
        """Close interface to queue."""


class Pub(RawQueue):
    """Publisher queue."""

    async def send_message(
        self,
        msg: bytes,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Send a message on a queue."""
        raise NotImplementedError()


class Sub(RawQueue):
    """Subscriber queue."""

    @property
    def prefetch(self) -> int:
        """Get prefetch."""
        return self._prefetch

    @prefetch.setter
    def prefetch(self, val: int) -> None:
        """Set prefetch."""
        if val < MIN_PREFETCH:
            raise ValueError(f"prefetch must be >= {MIN_PREFETCH}")
        self._prefetch = val

    @staticmethod
    def _to_message(*args: Any) -> Optional[Message]:
        """Convert broker_client-specific payload to standardized Message
        type."""
        raise NotImplementedError()

    async def get_message(
        self,
        timeout_millis: Optional[int],
        retries: int,
        retry_delay: float,
    ) -> Optional[Message]:
        """Get a single message from a queue."""
        raise NotImplementedError()

    async def ack_message(
        self,
        msg: Message,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Ack a message from the queue."""
        raise NotImplementedError()

    async def reject_message(
        self,
        msg: Message,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Reject (nack) a message from the queue."""
        raise NotImplementedError()

    def message_generator(  # NOTE: no `async` b/c it's abstract; overriding methods will need `async`
        self,
        timeout: int,
        propagate_error: bool,
        retries: int,
        retry_delay: float,
    ) -> AsyncGenerator[Optional[Message], None]:
        """Yield Messages.

        Asynchronously generate messages with variable timeout.
        Yield `None` on `athrow()`.

        Keyword Arguments:
            timeout {int} -- timeout in seconds for inactivity (default: {60})
            propagate_error {bool} -- should errors from downstream code kill the generator? (default: {True})
        """
        raise NotImplementedError()


class BrokerClient:
    """BrokerClient Pub-Sub Factory."""

    NAME = "abstract-broker_client"

    @staticmethod
    async def create_pub_queue(
        address: str,
        name: str,
        auth_token: str,
    ) -> Pub:
        """Create a publishing queue."""
        raise NotImplementedError()

    @staticmethod
    async def create_sub_queue(
        address: str,
        name: str,
        prefetch: int,
        auth_token: str,
    ) -> Sub:
        """Create a subscription queue."""
        raise NotImplementedError()



"""Unit test Queue class."""

# pylint:disable=invalid-name,protected-access

from typing import Any, AsyncGenerator
from unittest.mock import call, patch, sentinel

import pytest
from mqclient.broker_client_interface import (
    AckException,
    Message,
    MQClientException,
    NackException,
)
from mqclient.config import DEFAULT_RETRIES, DEFAULT_RETRY_DELAY
from mqclient.queue import EmptyQueueException, Queue

try:
    from unittest.mock import AsyncMock
except ImportError:
    from mock import AsyncMock  # use backport


@pytest.mark.asyncio
async def test_send() -> None:
    """Test send."""
    mock_broker_client = AsyncMock()
    with patch(
        "mqclient.broker_client_manager.get_broker_client"
    ) as mock_get_broker_client:
        mock_get_broker_client.return_value = mock_broker_client
        q = Queue("mock")

    data = {"a": 1234}
    async with q.open_pub() as p:
        await p.send(data)
    mock_broker_client.create_pub_queue.return_value.send_message.assert_awaited()
    mock_broker_client.create_pub_queue.return_value.close.assert_called()

    # send() adds a unique header, so we need to look at only the data
    msg = Message(
        id(sentinel.ID),
        mock_broker_client.create_pub_queue.return_value.send_message.call_args.args[0],
    )
    assert msg.data == data


@pytest.mark.asyncio
async def test_open_sub() -> None:
    """Test recv."""

    # pylint:disable=unused-argument
    async def gen(*args: Any, **kwargs: Any) -> AsyncGenerator[Message, None]:
        for i, d in enumerate(data):
            yield Message(i, Message.serialize(d))

    mock_broker_client = AsyncMock()
    with patch(
        "mqclient.broker_client_manager.get_broker_client"
    ) as mock_get_broker_client:
        mock_get_broker_client.return_value = mock_broker_client
        q = Queue("mock")

    data = ["a", {"b": 100}, ["foo", "bar"]]
    mock_broker_client.create_sub_queue.return_value.message_generator = gen

    async with q.open_sub() as stream:
        recv_data = [d async for d in stream]
        assert data == recv_data
        if not stream._sub:
            raise MQClientException("_sub not instantiated")
        stream._sub.ack_message.assert_has_calls(  # type: ignore[attr-defined]
            [call(Message(i, Message.serialize(d))) for i, d in enumerate(recv_data)]
        )

    mock_broker_client.create_sub_queue.return_value.close.assert_called()


@pytest.mark.asyncio
async def test_open_sub_one() -> None:
    """Test open_sub_one."""
    mock_broker_client = AsyncMock()
    with patch(
        "mqclient.broker_client_manager.get_broker_client"
    ) as mock_get_broker_client:
        mock_get_broker_client.return_value = mock_broker_client
        q = Queue("mock")

    data = {"b": 100}
    msg = Message(0, Message.serialize(data))
    mock_broker_client.create_sub_queue.return_value.get_message.return_value = msg

    async with q.open_sub_one() as d:
        recv_data = d

    assert data == recv_data
    mock_broker_client.create_sub_queue.return_value.ack_message.assert_called_with(
        msg,
        retries=DEFAULT_RETRIES,
        retry_delay=DEFAULT_RETRY_DELAY,
    )
    mock_broker_client.create_sub_queue.return_value.close.assert_called()


@pytest.mark.asyncio
async def test_open_sub_one__no_msg() -> None:
    """Test open_sub_one with an empty queue."""
    mock_broker_client = AsyncMock()
    with patch(
        "mqclient.broker_client_manager.get_broker_client"
    ) as mock_get_broker_client:
        mock_get_broker_client.return_value = mock_broker_client
        q = Queue("mock")

    mock_broker_client.create_sub_queue.return_value.get_message.return_value = None

    with pytest.raises(EmptyQueueException):
        async with q.open_sub_one() as _:
            assert 0  # we should never get here

    mock_broker_client.create_sub_queue.return_value.ack_message.assert_not_called()
    mock_broker_client.create_sub_queue.return_value.reject_message.assert_not_called()
    mock_broker_client.create_sub_queue.return_value.close.assert_called()


@pytest.mark.asyncio
async def test_safe_ack() -> None:
    """Test _safe_ack()."""
    mock_broker_client = AsyncMock()
    with patch(
        "mqclient.broker_client_manager.get_broker_client"
    ) as mock_get_broker_client:
        mock_get_broker_client.return_value = mock_broker_client
        q = Queue("mock")

    data = {"b": 100}

    # okay/normal
    mock_sub = AsyncMock()
    msg = Message(0, Message.serialize(data))
    assert msg._ack_status == Message.AckStatus.NONE
    await q._safe_ack(mock_sub, msg)
    mock_sub.ack_message.assert_called_with(
        msg,
        retries=DEFAULT_RETRIES,
        retry_delay=DEFAULT_RETRY_DELAY,
    )
    assert msg._ack_status == Message.AckStatus.ACKED

    # okay but pointless
    mock_sub = AsyncMock()
    msg = Message(0, Message.serialize(data))
    msg._ack_status = Message.AckStatus.ACKED
    assert msg._ack_status == Message.AckStatus.ACKED
    await q._safe_ack(mock_sub, msg)
    mock_sub.ack_message.assert_not_called()
    assert msg._ack_status == Message.AckStatus.ACKED

    # not okay
    mock_sub = AsyncMock()
    msg = Message(0, Message.serialize(data))
    msg._ack_status = Message.AckStatus.NACKED
    assert msg._ack_status == Message.AckStatus.NACKED
    with pytest.raises(AckException):
        await q._safe_ack(mock_sub, msg)
    mock_sub.ack_message.assert_not_called()
    assert msg._ack_status == Message.AckStatus.NACKED


@pytest.mark.asyncio
async def test_safe_nack() -> None:
    """Test _safe_nack()."""
    mock_broker_client = AsyncMock()
    with patch(
        "mqclient.broker_client_manager.get_broker_client"
    ) as mock_get_broker_client:
        mock_get_broker_client.return_value = mock_broker_client
        q = Queue("mock")

    data = {"b": 100}

    # okay/normal
    mock_sub = AsyncMock()
    msg = Message(0, Message.serialize(data))
    assert msg._ack_status == Message.AckStatus.NONE
    await q._safe_nack(mock_sub, msg)
    mock_sub.reject_message.assert_called_with(
        msg,
        retries=DEFAULT_RETRIES,
        retry_delay=DEFAULT_RETRY_DELAY,
    )
    assert msg._ack_status == Message.AckStatus.NACKED

    # not okay
    mock_sub = AsyncMock()
    msg = Message(0, Message.serialize(data))
    msg._ack_status = Message.AckStatus.ACKED
    assert msg._ack_status == Message.AckStatus.ACKED
    with pytest.raises(NackException):
        await q._safe_nack(mock_sub, msg)
    mock_sub.reject_message.assert_not_called()
    assert msg._ack_status == Message.AckStatus.ACKED

    # okay but pointless
    mock_sub = AsyncMock()
    msg = Message(0, Message.serialize(data))
    msg._ack_status = Message.AckStatus.NACKED
    assert msg._ack_status == Message.AckStatus.NACKED
    await q._safe_nack(mock_sub, msg)
    mock_sub.reject_message.assert_not_called()
    assert msg._ack_status == Message.AckStatus.NACKED


@pytest.mark.asyncio
async def test_nack_current() -> None:
    """Test recv with nack_current()."""

    # pylint:disable=unused-argument
    async def gen(*args: Any, **kwargs: Any) -> AsyncGenerator[Message, None]:
        for i, d in enumerate(data):
            yield Message(i, Message.serialize(d))

    mock_broker_client = AsyncMock()
    with patch(
        "mqclient.broker_client_manager.get_broker_client"
    ) as mock_get_broker_client:
        mock_get_broker_client.return_value = mock_broker_client
        q = Queue("mock")

    data = ["a", {"b": 100}, ["foo", "bar"]]
    msgs = [Message(i, Message.serialize(d)) for i, d in enumerate(data)]
    mock_broker_client.create_sub_queue.return_value.message_generator = gen

    async with q.open_sub() as stream:
        i = 0
        if not stream._sub:
            raise MQClientException("_sub not instantiated")
        # manual nacking won't actually place the message for redelivery b/c of mocking
        async for _ in stream:
            if i == 0:  # nack it
                await stream.nack_current()
                stream._sub.reject_message.assert_has_calls(  # type: ignore[attr-defined]
                    [
                        call(
                            msgs[0],
                            retries=DEFAULT_RETRIES,
                            retry_delay=DEFAULT_RETRY_DELAY,
                        )
                    ]
                )
            elif i == 1:  # DON'T nack it
                stream._sub.ack_message.assert_not_called()  # type: ignore[attr-defined]  # from i=0
            elif i == 2:  # nack it
                stream._sub.reject_message.assert_has_calls(  # type: ignore[attr-defined]
                    [
                        call(
                            msgs[0],
                            retries=DEFAULT_RETRIES,
                            retry_delay=DEFAULT_RETRY_DELAY,
                        )
                    ]
                )
                stream._sub.ack_message.assert_has_calls(  # type: ignore[attr-defined]
                    [
                        call(
                            msgs[1],
                            retries=DEFAULT_RETRIES,
                            retry_delay=DEFAULT_RETRY_DELAY,
                        )
                    ]
                )
                await stream.nack_current()
                stream._sub.reject_message.assert_has_calls(  # type: ignore[attr-defined]
                    [
                        call(
                            msgs[0],
                            retries=DEFAULT_RETRIES,
                            retry_delay=DEFAULT_RETRY_DELAY,
                        ),
                        call(
                            msgs[2],
                            retries=DEFAULT_RETRIES,
                            retry_delay=DEFAULT_RETRY_DELAY,
                        ),
                    ]
                )
            else:
                assert 0
            i += 1
        stream._sub.ack_message.assert_has_calls(  # type: ignore[attr-defined]
            [
                call(
                    msgs[1],
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
            ]
        )
        stream._sub.reject_message.assert_has_calls(  # type: ignore[attr-defined]
            [
                call(
                    msgs[0],
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                ),
                call(
                    msgs[2],
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                ),
            ]
        )



"""A worker processes messages from one queue, and sends results on a second
queue."""

import argparse
import asyncio
import logging
import subprocess

import coloredlogs  # type: ignore[import]
from mqclient import Queue


async def worker(recv_queue: Queue, send_queue: Queue) -> None:
    """Demo example worker."""
    async with recv_queue.open_sub() as stream, send_queue.open_pub() as p:
        async for data in stream:
            cmd = data["cmd"]
            out = subprocess.check_output(cmd, shell=True)
            data["out"] = out.decode("utf-8")
            await p.send(data)


if __name__ == "__main__":
    coloredlogs.install(level=logging.DEBUG)

    parser = argparse.ArgumentParser(description="Worker")
    parser.add_argument("--broker-client", required=True, help="which kind of broker")
    parser.add_argument("--address", default="localhost", help="queue address")
    parser.add_argument("--in-queue", default="queue1", help="input queue")
    parser.add_argument("--out-queue", default="queue2", help="output queue")
    parser.add_argument("--prefetch", type=int, default=10, help="input queue prefetch")
    args = parser.parse_args()

    inq = Queue(
        args.broker_client,
        address=args.address,
        name=args.in_queue,
        prefetch=args.prefetch,
    )
    outq = Queue(
        args.broker_client,
        address=args.address,
        name=args.out_queue,
    )

    asyncio.get_event_loop().run_until_complete(worker(inq, outq))



"""Unit test the broker client manager."""

import re

import pytest
from mqclient import broker_client_manager
from mqclient.broker_client_interface import MQClientException


def test_missing_broker_clients() -> None:
    """Test legitimate, but not-installed broker clients."""
    for name in ["pulsar", "rabbitmq", "nats"]:
        with pytest.raises(
            MQClientException,
            match=re.escape(
                f"Install the '{name}' extra if you want to use the '{name}' broker client"
            ),
        ):
            broker_client_manager.get_broker_client(name)


def test_invalid_broker_clients() -> None:
    """Test illegitimate broker clients."""
    for name in ["foo", "bar", "baz"]:
        with pytest.raises(
            MQClientException,
            match=re.escape(f"Unknown broker client: {name}"),
        ):
            broker_client_manager.get_broker_client(name)



"""Run integration tests for NATS broker_client."""

import asyncio
import logging

import pytest
from mqclient import broker_client_manager

from ..abstract_broker_client_tests import (
    integrate_broker_client_interface,
    integrate_queue,
)
from ..abstract_broker_client_tests.utils import (  # pytest.fixture # noqa: F401 # pylint: disable=W0611
    queue_name,
)

logging.getLogger().setLevel(logging.DEBUG)
logging.getLogger("flake8").setLevel(logging.WARNING)


@pytest.fixture(scope="module")
def event_loop():  # type: ignore[no-untyped-def]
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


class TestNATSQueue(integrate_queue.PubSubQueue):
    """Run PubSubQueue integration tests with NATS broker_client."""

    broker_client = "nats"


class TestNATSBrokerClient(
    integrate_broker_client_interface.PubSubBrokerClientInterface
):
    """Run PubSubBrokerClientInterface integration tests with NATS broker_client."""

    broker_client = broker_client_manager.get_broker_client("nats")



"""Run integration tests for given broker_client, on Queue class."""

# pylint:disable=invalid-name,too-many-public-methods,redefined-outer-name,unused-import

import asyncio
import logging
import random
from multiprocessing.dummy import Pool as ThreadPool
from typing import Any, List, Optional
from unittest.mock import patch

import asyncstdlib as asl
import pytest
from mqclient.broker_client_interface import MQClientException
from mqclient.queue import EmptyQueueException, Queue

from .utils import (
    DATA_LIST,
    _log_recv,
    _log_recv_multiple,
    _log_send,
    all_were_received,
)

#
# retry stuff
#


CI_TEST_RETRY_TRIGGER = "mqclient.broker_clients.utils._ci_test_retry_trigger"


class FailFirstTryException(Exception):
    pass


def fail_first_try(attempt: int) -> None:
    if attempt == 0:
        raise FailFirstTryException()


#
# tests
#


PREFETCH_TEST_VALUES = [None, 1, 2, len(DATA_LIST), 50]


class PubSubQueue:
    """Integration test suite for Queue objects."""

    broker_client: str = ""

    ###########################################################################
    # tests 000 - 099:
    #
    # Testing scenarios with different numbers of sub and/or pubs
    # to see no data loss
    ###########################################################################

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_000(self, queue_name: str, auth_token: str) -> None:
        """Test one pub, one sub."""
        all_recvd: List[Any] = []

        pub_sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        async with pub_sub.open_pub() as p:
            await p.send(DATA_LIST[0])
            _log_send(DATA_LIST[0])

        async with pub_sub.open_sub_one() as d:
            all_recvd.append(_log_recv(d))
            assert d == DATA_LIST[0]

        async with pub_sub.open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        pub_sub.timeout = 1
        async with pub_sub.open_sub() as gen:
            async for i, d in asl.enumerate(gen):
                print(f"{i}: `{d}`")
                all_recvd.append(_log_recv(d))
                # assert d == DATA_LIST[i]  # we don't guarantee order

        assert all_were_received(all_recvd, [DATA_LIST[0]] + DATA_LIST)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_001(self, queue_name: str, auth_token: str) -> None:
        """Test an individual pub and an individual sub."""
        all_recvd: List[Any] = []

        pub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        async with pub.open_pub() as p:
            await p.send(DATA_LIST[0])
            _log_send(DATA_LIST[0])

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        async with sub.open_sub_one() as d:
            all_recvd.append(_log_recv(d))
            assert d == DATA_LIST[0]

        async with pub.open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        sub.timeout = 1
        async with sub.open_sub() as gen:
            async for i, d in asl.enumerate(gen):
                print(f"{i}: `{d}`")
                all_recvd.append(_log_recv(d))
                # assert d == DATA_LIST[i]  # we don't guarantee order

        assert all_were_received(all_recvd, [DATA_LIST[0]] + DATA_LIST)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_002(self, queue_name: str, auth_token: str) -> None:
        """Failure-test one pub, two subs (one subscribed to wrong queue)."""
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            await p.send(DATA_LIST[0])
            _log_send(DATA_LIST[0])

        with pytest.raises(Exception):
            name = f"{queue_name}-fail"
            async with Queue(self.broker_client, name=name).open_sub_one() as d:
                all_recvd.append(_log_recv(d))

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_sub_one() as d:
            all_recvd.append(_log_recv(d))
            assert d == DATA_LIST[0]

        assert all_were_received(all_recvd, [DATA_LIST[0]])

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_010(self, queue_name: str, auth_token: str) -> None:
        """Test one pub, multiple subs, ordered/alternatingly."""
        all_recvd: List[Any] = []

        # for each send, create and receive message via a new sub
        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for data in DATA_LIST:
                await p.send(data)
                _log_send(data)

                async with Queue(
                    self.broker_client, name=queue_name, auth_token=auth_token
                ).open_sub_one() as d:
                    all_recvd.append(_log_recv(d))
                    assert d == data

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    @pytest.mark.parametrize(
        "num_subs",
        [
            len(DATA_LIST) // 2,
            len(DATA_LIST),
            len(DATA_LIST) ** 2,
        ],
    )
    async def test_020(self, queue_name: str, auth_token: str, num_subs: int) -> None:
        """Test one pub, multiple subs, unordered (front-loaded sending).

        Uses `open_sub()`
        """
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for data in DATA_LIST:
                await p.send(data)
                _log_send(data)

        subs = []
        for _ in range(num_subs):
            subs.append(
                Queue(
                    self.broker_client,
                    name=queue_name,
                    auth_token=auth_token,
                    timeout=1,
                )
            )
            await asyncio.sleep(0.1)

        for i, sub in enumerate(subs):
            async with sub.open_sub() as gen:
                recv_data_list = [_log_recv(m) async for m in gen]
                if i < len(DATA_LIST):
                    assert recv_data_list
                else:
                    assert not recv_data_list
                all_recvd.extend(recv_data_list)

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    @pytest.mark.parametrize(
        "num_subs",
        [
            len(DATA_LIST) // 2,
            len(DATA_LIST),
            len(DATA_LIST) ** 2,
        ],
    )
    async def test_021__threaded(
        self, queue_name: str, auth_token: str, num_subs: int
    ) -> None:
        """Test one pub, multiple subs, unordered (front-loaded sending).

        Uses `open_sub()`
        """
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for data in DATA_LIST:
                await p.send(data)
                _log_send(data)

        async def recv_thread(i: int) -> List[Any]:
            sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
            sub.timeout = 1
            async with sub.open_sub() as gen:
                recv_data_list = [m async for m in gen]
            return _log_recv_multiple(recv_data_list)

        def start_recv_thread(num_id: int) -> Any:
            return asyncio.run(recv_thread(num_id))

        with ThreadPool(num_subs) as pool:
            received_data = pool.map(start_recv_thread, range(num_subs))

        n_subs_that_got_msgs = 0
        for sublist in received_data:
            if sublist:
                n_subs_that_got_msgs += 1
                all_recvd.extend(sublist)
        # since threads are mixed, can't test like test_020
        # the threading makes us not able to assert how many, but it should be >= 1
        logging.debug(f"{n_subs_that_got_msgs=}")
        assert n_subs_that_got_msgs > 2

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_030(self, queue_name: str, auth_token: str) -> None:
        """Test one pub, multiple subs, unordered (front-loaded sending).

        Use the same number of subs as number of messages.

        Uses `open_sub_one()`
        """
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for data in DATA_LIST:
                await p.send(data)
                _log_send(data)

        subs = [
            Queue(
                self.broker_client,
                name=queue_name,
                auth_token=auth_token,
                timeout=1,
            )
            for _ in range(len(DATA_LIST))
        ]

        for sub in subs:
            async with sub.open_sub_one() as m:
                all_recvd.append(_log_recv(m))

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_031(self, queue_name: str, auth_token: str) -> None:
        """Failure-test one pub, and too many subs.

        More subs than messages with `open_sub_one()` will raise an
        exception.

        Uses `open_sub_one()`
        """
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for data in DATA_LIST:
                await p.send(data)
                _log_send(data)

        subs = [
            Queue(
                self.broker_client,
                name=queue_name,
                auth_token=auth_token,
                timeout=1,
            )
            for _ in range(len(DATA_LIST) + 1)
        ]

        for i, sub in enumerate(subs):
            if i == len(DATA_LIST):
                with pytest.raises(EmptyQueueException):
                    async with sub.open_sub_one() as m:
                        all_recvd.append(_log_recv(m))
            else:
                async with sub.open_sub_one() as m:
                    all_recvd.append(_log_recv(m))

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_032__threaded(self, queue_name: str, auth_token: str) -> None:
        """Test one pub, multiple subs, unordered (front-loaded sending).

        Use the same number of subs as number of messages.

        Uses `open_sub_one()`
        """
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for data in DATA_LIST:
                await p.send(data)
                _log_send(data)

        async def recv_thread(_: int) -> Any:
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_sub_one() as d:
                recv_data = d
            return _log_recv(recv_data)

        def start_recv_thread(num_id: int) -> Any:
            return asyncio.run(recv_thread(num_id))

        with ThreadPool(len(DATA_LIST)) as pool:
            all_recvd = pool.map(start_recv_thread, range(len(DATA_LIST)))

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_033__threaded(self, queue_name: str, auth_token: str) -> None:
        """Failure-test one pub, and too many subs.

        More subs than messages with `open_sub_one()` will raise an
        exception.

        Uses `open_sub_one()`
        """
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for data in DATA_LIST:
                await p.send(data)
                _log_send(data)

        async def recv_thread(_: int) -> Any:
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_sub_one() as d:
                recv_data = d
            return _log_recv(recv_data)

        def start_recv_thread(num_id: int) -> Any:
            return asyncio.run(recv_thread(num_id))

        with ThreadPool(len(DATA_LIST)) as pool:
            all_recvd = pool.map(start_recv_thread, range(len(DATA_LIST)))

        # Extra Sub
        with pytest.raises(EmptyQueueException):
            await recv_thread(-1)

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_060(self, queue_name: str, auth_token: str) -> None:
        """Test multiple pubs, one sub, ordered/alternatingly."""
        all_recvd: List[Any] = []

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)

        for data in DATA_LIST:
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_pub() as p:
                await p.send(data)
                _log_send(data)

            sub.timeout = 1
            sub.except_errors = False
            async with sub.open_sub() as gen:
                received_data = [m async for m in gen]
            all_recvd.extend(_log_recv_multiple(received_data))

            assert len(received_data) == 1
            assert data == received_data[0]

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_061(self, queue_name: str, auth_token: str) -> None:
        """Test multiple pubs, one sub, unordered (front-loaded sending)."""
        all_recvd: List[Any] = []

        for data in DATA_LIST:
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_pub() as p:
                await p.send(data)
                _log_send(data)

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        sub.timeout = 1
        async with sub.open_sub() as gen:
            received_data = [m async for m in gen]
        all_recvd.extend(_log_recv_multiple(received_data))

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_080(self, queue_name: str, auth_token: str) -> None:
        """Test multiple pubs, multiple subs, ordered/alternatingly.

        Use the same number of pubs as subs.
        """
        all_recvd: List[Any] = []

        for data in DATA_LIST:
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_pub() as p:
                await p.send(data)
                _log_send(data)

            sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
            sub.timeout = 1
            async with sub.open_sub() as gen:
                received_data = [m async for m in gen]
            assert received_data
            all_recvd.extend(_log_recv_multiple(received_data))

            assert len(received_data) == 1
            assert data == received_data[0]

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_081(self, queue_name: str, auth_token: str) -> None:
        """Test multiple pubs, multiple subs, unordered (front-loaded sending).

        Use the same number of pubs as subs.
        """
        all_recvd: List[Any] = []

        for data in DATA_LIST:
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_pub() as p:
                await p.send(data)
                _log_send(data)

        for _ in range(len(DATA_LIST)):
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_sub_one() as d:
                all_recvd.append(_log_recv(d))

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_082(self, queue_name: str, auth_token: str) -> None:
        """Test multiple pubs, multiple subs, unordered (front-loaded sending).

        Use the more pubs than subs.
        """
        all_recvd: List[Any] = []

        for data in DATA_LIST:
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_pub() as p:
                await p.send(data)
                _log_send(data)

        for i in range(len(DATA_LIST)):
            if i % 2 == 0:  # each sub receives 2 messages back-to-back
                sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
            async with sub.open_sub_one() as d:
                all_recvd.append(_log_recv(d))

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_083(self, queue_name: str, auth_token: str) -> None:
        """Test multiple pubs, multiple subs, unordered (front-loaded sending).

        Use the fewer pubs than subs.
        """
        all_recvd: List[Any] = []

        for data_pairs in [DATA_LIST[i : i + 2] for i in range(0, len(DATA_LIST), 2)]:
            for data in data_pairs:
                async with Queue(
                    self.broker_client, name=queue_name, auth_token=auth_token
                ).open_pub() as p:
                    await p.send(data)
                    _log_send(data)

        for _ in range(len(DATA_LIST)):
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_sub_one() as d:
                all_recvd.append(_log_recv(d))

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_090(self, queue_name: str, auth_token: str) -> None:
        """Test_20 with variable prefetching.

        One pub, multiple subs.
        """
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for i in range(1, len(DATA_LIST) * 2):
                # for each send, create and receive message via a new sub
                for data in DATA_LIST:
                    await p.send(data)
                    _log_send(data)

                    sub = Queue(
                        self.broker_client,
                        name=queue_name,
                        auth_token=auth_token,
                        prefetch=i,
                    )
                    async with sub.open_sub_one() as d:
                        all_recvd.append(_log_recv(d))
                        assert d == data

        assert all_were_received(all_recvd, DATA_LIST * ((len(DATA_LIST) * 2) - 1))

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_091(self, queue_name: str, auth_token: str) -> None:
        """Test one pub, multiple subs, with prefetching.

        Prefetching should have no visible affect.
        """
        all_recvd: List[Any] = []

        for data in DATA_LIST:
            async with Queue(
                self.broker_client, name=queue_name, auth_token=auth_token
            ).open_pub() as p:
                await p.send(data)
                _log_send(data)

        # this should not eat up the whole queue
        sub = Queue(
            self.broker_client, name=queue_name, auth_token=auth_token, prefetch=20
        )
        async with sub.open_sub_one() as d:
            all_recvd.append(_log_recv(d))
        async with sub.open_sub_one() as d:
            all_recvd.append(_log_recv(d))

        sub2 = Queue(
            self.broker_client, name=queue_name, auth_token=auth_token, prefetch=2
        )
        sub2.timeout = 1
        async with sub2.open_sub() as gen:
            async for _, d in asl.enumerate(gen):
                all_recvd.append(_log_recv(d))

        assert all_were_received(all_recvd)

    ###########################################################################
    # tests 100 - 199:
    #
    # Tests for open_sub()
    ###########################################################################

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_100(self, queue_name: str, auth_token: str) -> None:
        """Test open_sub() fail and recovery, with multiple open_sub()
        calls."""
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        class TestException(Exception):  # pylint: disable=C0115
            pass

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        sub.timeout = 1
        async with sub.open_sub() as gen:
            async for i, d in asl.enumerate(gen):
                print(f"{i}: `{d}`")
                if i == 2:
                    raise TestException()
                all_recvd.append(_log_recv(d))
                # assert d == DATA_LIST[i]  # we don't guarantee order

        logging.warning("Round 2!")

        # continue where we left off
        reused = False
        sub.timeout = 1
        async with sub.open_sub() as gen:
            async for i, d in asl.enumerate(gen):
                print(f"{i}: `{d}`")
                reused = True
                all_recvd.append(_log_recv(d))
                # assert d == DATA_LIST[i]  # we don't guarantee order
        assert reused
        print(all_recvd)
        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_101(self, queue_name: str, auth_token: str) -> None:
        """Test open_sub() fail and recovery, with error propagation."""
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        class TestException(Exception):  # pylint: disable=C0115
            pass

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        excepted = False
        try:
            sub.timeout = 1
            sub.except_errors = False
            async with sub.open_sub() as gen:
                async for i, d in asl.enumerate(gen):
                    if i == 2:
                        raise TestException()
                    all_recvd.append(_log_recv(d))
                    # assert d == DATA_LIST[i]  # we don't guarantee order
        except TestException:
            excepted = True
        assert excepted

        logging.warning("Round 2!")

        # continue where we left off
        reused = False
        sub.timeout = 1
        sub.except_errors = False
        async with sub.open_sub() as gen:
            async for i, d in asl.enumerate(gen):
                reused = True
                all_recvd.append(_log_recv(d))
                # assert d == DATA_LIST[i]  # we don't guarantee order
        assert reused

        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_110__fail(self, queue_name: str, auth_token: str) -> None:
        """Failure-test open_sub() with reusing a 'QueueSubResource'
        instance."""
        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        sub.timeout = 1
        recv_gen = sub.open_sub()
        async with recv_gen as gen:
            async for i, d in asl.enumerate(gen):
                print(f"{i}: `{d}`")
                # assert d == DATA_LIST[i]  # we don't guarantee order

        logging.warning("Round 2!")

        # continue where we left off
        with pytest.raises(MQClientException):
            async with recv_gen as gen:
                assert 0  # we should never get here

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_120_break(self, queue_name: str, auth_token: str) -> None:
        """Test open_sub() with a `break` statement."""
        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        sub.timeout = 1
        all_recvd = []
        async with sub.open_sub() as gen:
            async for i, d in asl.enumerate(gen):
                print(f"{i}: `{d}`")
                all_recvd.append(_log_recv(d))
                if i == 2:
                    break  # NOTE: break is treated as a good exit, so the msg is acked

        logging.warning("Round 2!")

        # continue where we left off
        async with sub.open_sub() as gen:
            async for i, d in asl.enumerate(gen):
                print(f"{i}: `{d}`")
                all_recvd.append(_log_recv(d))

        assert all_were_received(all_recvd)

    ###########################################################################
    # tests 200 - 299:
    #
    # Tests for open_sub_manual_acking()
    ###########################################################################

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    @pytest.mark.parametrize("sub_queue_prefetch", PREFETCH_TEST_VALUES)
    async def test_200__ideal(
        self,
        queue_name: str,
        auth_token: str,
        sub_queue_prefetch: Optional[int],
    ) -> None:
        """Test open_sub_manual_acking() ideal scenario."""
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        if sub_queue_prefetch is not None:
            sub = Queue(
                self.broker_client,
                name=queue_name,
                auth_token=auth_token,
                prefetch=sub_queue_prefetch,
            )
        else:
            sub = Queue(
                self.broker_client,
                name=queue_name,
                auth_token=auth_token,
            )
        sub.timeout = 1
        async with sub.open_sub_manual_acking() as gen:
            async for i, msg in asl.enumerate(gen.iter_messages()):
                print(f"{i}: `{msg.data}`")
                all_recvd.append(_log_recv(msg.data))
                # assert msg.data == DATA_LIST[i]  # we don't guarantee order
                await gen.ack(msg)

        print(all_recvd)
        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    @pytest.mark.parametrize("sub_queue_prefetch", PREFETCH_TEST_VALUES)
    async def test_202__delayed_mixed_acking_nacking(
        self,
        queue_name: str,
        auth_token: str,
        sub_queue_prefetch: Optional[int],
    ) -> None:
        """Test open_sub_manual_acking() fail and immediate recovery with
        multi-tasking, with mixed acking and nacking."""
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        class TestException(Exception):  # pylint: disable=C0115
            pass

        if sub_queue_prefetch is not None:
            sub = Queue(
                self.broker_client,
                name=queue_name,
                auth_token=auth_token,
                prefetch=sub_queue_prefetch,
            )
        else:
            sub = Queue(
                self.broker_client,
                name=queue_name,
                auth_token=auth_token,
            )
        sub.timeout = 1
        async with sub.open_sub_manual_acking() as gen:
            pending = []
            async for i, msg in asl.enumerate(gen.iter_messages()):
                try:
                    # DO WORK!
                    print(f"{i}: `{msg.data}`")
                    if i % 3 == 0:  # nack every 1/3
                        raise TestException()
                    all_recvd.append(_log_recv(msg.data))
                    pending.append(msg)
                    # assert msg.data == DATA_LIST[i]  # we don't guarantee order
                    if i % 2 == 0:  # ack every 1/2
                        print(f"ack {i}: `{msg.data}`")
                        await gen.ack(msg)
                        pending.remove(msg)
                except Exception:
                    print(f"nack {i}: `{msg.data}`")
                    await gen.nack(msg)

            for msg in pending:  # messages with index not %2 nor %3, (1,5,7,...)
                await gen.ack(msg)

        print(all_recvd)
        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    @pytest.mark.parametrize("sub_queue_prefetch", PREFETCH_TEST_VALUES)
    async def test_204__post_ack(
        self,
        queue_name: str,
        auth_token: str,
        sub_queue_prefetch: Optional[int],
    ) -> None:
        """Test open_sub_manual_acking() where messages aren't acked until
        after all have been received."""
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        if sub_queue_prefetch is not None:
            sub = Queue(
                self.broker_client,
                name=queue_name,
                auth_token=auth_token,
                prefetch=sub_queue_prefetch,
            )
        else:
            sub = Queue(
                self.broker_client,
                name=queue_name,
                auth_token=auth_token,
            )
        sub.timeout = 1
        to_ack = []
        async with sub.open_sub_manual_acking() as gen:
            async for i, msg in asl.enumerate(gen.iter_messages()):
                print(f"{i}: `{msg.data}`")
                all_recvd.append(_log_recv(msg.data))
                to_ack.append(msg)
                # assert msg.data == DATA_LIST[i]  # we don't guarantee order

            iter_em = list(enumerate(to_ack))
            random.shuffle(iter_em)
            for i, msg in iter_em:
                print(f"ack {i} (shuffled): `{msg.data}`")
                await gen.ack(msg)

        print(all_recvd)
        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_210__immediate_recovery(
        self,
        queue_name: str,
        auth_token: str,
    ) -> None:
        """Test open_sub_manual_acking() fail and immediate recovery, with
        nacking."""
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        class TestException(Exception):  # pylint: disable=C0115
            pass

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        sub.timeout = 1
        async with sub.open_sub_manual_acking() as gen:
            async for i, msg in asl.enumerate(gen.iter_messages()):
                try:
                    # DO WORK!
                    print(f"{i}: `{msg.data}`")
                    if i == 2:
                        raise TestException()
                    all_recvd.append(_log_recv(msg.data))
                    # assert msg.data == DATA_LIST[i]  # we don't guarantee order
                except Exception:
                    await gen.nack(msg)
                else:
                    await gen.ack(msg)

        print(all_recvd)
        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_220__posthoc_recovery(
        self,
        queue_name: str,
        auth_token: str,
    ) -> None:
        """Test open_sub_manual_acking() fail and post-hoc recovery, with
        nacking."""
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        class TestException(Exception):  # pylint: disable=C0115
            pass

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        excepted = False
        sub.timeout = 1
        # sub.except_errors = False  # has no effect
        async with sub.open_sub_manual_acking() as gen:
            try:
                async for i, msg in asl.enumerate(gen.iter_messages()):
                    print(f"{i}: `{msg.data}`")
                    if i == 2:
                        raise TestException()
                    all_recvd.append(_log_recv(msg.data))
                    # assert msg.data == DATA_LIST[i]  # we don't guarantee order
                    await gen.ack(msg)
            except TestException:
                excepted = True
                await gen.nack(msg)
        assert excepted

        logging.warning("Round 2!")

        # continue where we left off
        posthoc = False
        sub.timeout = 1
        async with sub.open_sub_manual_acking() as gen:
            async for i, msg in asl.enumerate(gen.iter_messages()):
                print(f"{i}: `{msg.data}`")
                posthoc = True
                all_recvd.append(_log_recv(msg.data))
                # assert msg.data == DATA_LIST[i]  # we don't guarantee order
                await gen.ack(msg)
        assert posthoc
        print(all_recvd)
        assert all_were_received(all_recvd)

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_221__posthoc_recovery__fail(
        self,
        queue_name: str,
        auth_token: str,
    ) -> None:
        """Test open_sub_manual_acking() fail, post-hoc recovery, then fail.

        Final fail is due to not nacking.
        """
        all_recvd: List[Any] = []

        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        class TestException(Exception):  # pylint: disable=C0115
            pass

        errored_msg = None

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        excepted = False
        async with sub.open_sub_manual_acking() as gen:
            try:
                async for i, msg in asl.enumerate(gen.iter_messages()):
                    print(f"{i}: `{msg.data}`")
                    if i == 2:
                        errored_msg = msg.data
                        raise TestException()
                    all_recvd.append(_log_recv(msg.data))
                    # assert msg.data == DATA_LIST[i]  # we don't guarantee order
                    await gen.ack(msg)
            except TestException:
                excepted = True
                # await gen.nack(msg)  # no acking
        assert excepted

        logging.warning("Round 2!")

        # continue where we left off
        posthoc = False
        sub.timeout = 1
        async with sub.open_sub_manual_acking() as gen:
            async for i, msg in asl.enumerate(gen.iter_messages()):
                print(f"{i}: `{msg.data}`")
                posthoc = True
                all_recvd.append(_log_recv(msg.data))
                # assert msg.data == DATA_LIST[i]  # we don't guarantee order
                await gen.ack(msg)
        assert posthoc

        # Either all the messages have been gotten (re-opening the connection took longer enough)
        # OR it hasn't been long enough to redeliver un-acked/nacked message
        # This is difficult to test -- all we can tell is if it is one of these scenarios
        print(all_recvd)
        assert all_were_received(all_recvd) or (
            all_were_received(all_recvd + [errored_msg])
        )

    @pytest.mark.asyncio
    @patch(CI_TEST_RETRY_TRIGGER, new=fail_first_try)
    async def test_230__fail_bad_usage(
        self,
        queue_name: str,
        auth_token: str,
    ) -> None:
        """Failure-test open_sub_manual_acking() with reusing a
        'QueueSubResource' instance."""
        async with Queue(
            self.broker_client, name=queue_name, auth_token=auth_token
        ).open_pub() as p:
            for d in DATA_LIST:
                await p.send(d)
                _log_send(d)

        sub = Queue(self.broker_client, name=queue_name, auth_token=auth_token)
        sub.timeout = 1
        recv_gen = sub.open_sub_manual_acking()
        async with recv_gen as gen:
            async for i, msg in asl.enumerate(gen.iter_messages()):
                print(f"{i}: `{msg.data}`")
                # assert msg.data == DATA_LIST[i]  # we don't guarantee order
                await gen.ack(msg)

        logging.warning("Round 2!")

        # continue where we left off
        with pytest.raises((AttributeError, RuntimeError)):
            # AttributeError: '_AsyncGeneratorContextManager' object has no attribute 'args'
            # RuntimeError: generator didn't yield
            async with recv_gen as gen:
                assert 0  # we should never get here



"""Unit Tests for RabbitMQ/Pika BrokerClient."""

import itertools
from typing import Any, List, Optional, Tuple
from unittest.mock import MagicMock

import pika  # type: ignore[import]
import pytest
from mqclient import broker_client_manager
from mqclient.broker_client_interface import Message, MQClientException
from mqclient.broker_clients.rabbitmq import HUMAN_PATTERN, _get_credentials, _parse_url
from mqclient.config import (
    DEFAULT_RETRIES,
    DEFAULT_RETRY_DELAY,
    DEFAULT_TIMEOUT,
    DEFAULT_TIMEOUT_MILLIS,
)

from ...abstract_broker_client_tests.unit_tests import BrokerClientUnitTest


class TestUnitRabbitMQ(BrokerClientUnitTest):
    """Unit test suite interface for RabbitMQ broker_client."""

    broker_client = broker_client_manager.get_broker_client("rabbitmq")
    con_patch = "pika.BlockingConnection"

    @staticmethod
    def _assert_nack_mock(mock_con: Any, called: bool, *with_args: Any) -> None:
        """Assert mock 'nack' function called (or not)."""
        if called:
            mock_con.return_value.channel.return_value.basic_nack.assert_called_with(
                *with_args,
                multiple=False,
                requeue=True,
            )
        else:
            mock_con.return_value.channel.return_value.basic_nack.assert_not_called()

    @staticmethod
    def _assert_ack_mock(mock_con: Any, called: bool, *with_args: Any) -> None:
        """Assert mock 'ack' function called (or not)."""
        if called:
            mock_con.return_value.channel.return_value.basic_ack.assert_called_with(
                *with_args,
                multiple=False,
            )
        else:
            mock_con.return_value.channel.return_value.basic_ack.assert_not_called()

    @staticmethod
    def _get_close_mock_fn(mock_con: Any) -> Any:
        """Return mock 'close' function call."""
        return mock_con.return_value.close

    @staticmethod
    async def _enqueue_mock_messages(
        mock_con: Any, data: List[bytes], ids: List[int], append_none: bool = True
    ) -> None:
        """Place messages on the mock queue."""
        if len(data) != len(ids):
            raise AttributeError("`data` and `ids` must have the same length.")
        messages = [(MagicMock(delivery_tag=i), None, d) for d, i in zip(data, ids)]
        if append_none:
            messages += [(None, None, None)]  # type: ignore
        mock_con.return_value.channel.return_value.consume.return_value.__next__.side_effect = (
            messages
        )

    @pytest.mark.asyncio
    async def test_create_pub_queue(self, mock_con: Any, queue_name: str) -> None:
        """Test creating pub queue."""
        pub = await self.broker_client.create_pub_queue("localhost", queue_name, "")
        assert pub.queue == queue_name  # type: ignore
        mock_con.return_value.channel.assert_called()

    @pytest.mark.asyncio
    async def test_create_sub_queue(self, mock_con: Any, queue_name: str) -> None:
        """Test creating sub queue."""
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 213, ""
        )
        assert sub.queue == queue_name  # type: ignore
        assert sub.prefetch == 213
        mock_con.return_value.channel.assert_called()

    @pytest.mark.asyncio
    async def test_send_message(self, mock_con: Any, queue_name: str) -> None:
        """Test sending message."""
        pub = await self.broker_client.create_pub_queue("localhost", queue_name, "")
        await pub.send_message(
            b"foo, bar, baz",
            retries=DEFAULT_RETRIES,
            retry_delay=DEFAULT_RETRY_DELAY,
        )
        mock_con.return_value.channel.return_value.basic_publish.assert_called_with(
            exchange="", routing_key=queue_name, body=b"foo, bar, baz"
        )

    @pytest.mark.asyncio
    async def test_get_message(self, mock_con: Any, queue_name: str) -> None:
        """Test getting message."""
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        mock_con.return_value.is_closed = False  # HACK - manually set attr

        fake_message = (MagicMock(delivery_tag=12), None, Message.serialize("foo, bar"))
        mock_con.return_value.channel.return_value.consume.return_value.__next__.side_effect = [
            fake_message
        ]
        m = await sub.get_message(
            timeout_millis=DEFAULT_TIMEOUT_MILLIS,
            retries=DEFAULT_RETRIES,
            retry_delay=DEFAULT_RETRY_DELAY,
        )
        assert m is not None
        assert m.msg_id == 12
        assert m.data == "foo, bar"

    @pytest.mark.asyncio
    async def test_message_generator_10_upstream_error(
        self, mock_con: Any, queue_name: str
    ) -> None:
        """Failure-test message generator.

        Generator should raise Exception originating upstream (a.k.a.
        from pika-package code).
        """
        sub = await self.broker_client.create_sub_queue("localhost", queue_name, 1, "")
        mock_con.return_value.is_closed = False  # HACK - manually set attr

        retries = 2  # >= 0

        class _MyException(Exception):
            pass

        mock_con.return_value.channel.return_value.consume.return_value.__next__.side_effect = (
            _MyException
        )
        with pytest.raises(_MyException):
            async for m in sub.message_generator(
                timeout=DEFAULT_TIMEOUT,
                propagate_error=True,
                retries=retries,
                retry_delay=DEFAULT_RETRY_DELAY,
            ):
                pass

        # would be called by Queue one more time
        assert self._get_close_mock_fn(mock_con).call_count == 0

        # reset for next call
        self._get_close_mock_fn(mock_con).reset_mock()

        # `propagate_error` attribute has no affect (b/c it deals w/ *downstream* errors)
        mock_con.return_value.channel.return_value.consume.return_value.__next__.side_effect = (
            _MyException
        )
        with pytest.raises(_MyException):
            async for m in sub.message_generator(
                timeout=DEFAULT_TIMEOUT,
                propagate_error=False,
                retries=retries,
                retry_delay=DEFAULT_RETRY_DELAY,
            ):
                pass

        # would be called by Queue one more time
        assert self._get_close_mock_fn(mock_con).call_count == 0


class TestUnitRabbitMQHelpers:
    """Unit test rabbitmq-specific helper functions."""

    def test_000(self) -> None:
        """Sanity check the constants."""
        assert HUMAN_PATTERN == ("[SCHEME://][USER[:PASS]@]HOST[:PORT][/VIRTUAL_HOST]")

    def test_100(self) -> None:
        """Test normal (successful) parsing of `_parse_url()`."""

        def _get_return_tuple(
            subdict: dict, password: Optional[str] = ""
        ) -> Tuple[dict, Optional[str], Optional[str]]:
            return (
                {k: v for k, v in subdict.items() if k not in ["username", "password"]},
                subdict.get("username", None),
                subdict.get("password", password),
            )

        tokens = dict(scheme="wxyz", port=1234, virtual_host="foo", username="hank")
        # test with every number of combinations of `tokens`
        for rlength in range(len(tokens) + 1):
            for _subset in itertools.combinations(tokens.items(), rlength):
                subdict = dict(_subset)

                # host is mandatory
                host = "localhost"
                subdict["host"] = host

                # optional tokens
                if user := subdict.get("username", ""):
                    user = f"{user}@"
                if port := subdict.get("port", ""):
                    port = f":{port}"
                if vhost := subdict.get("virtual_host", ""):
                    vhost = f"/{vhost}"
                if skm := subdict.get("scheme", ""):
                    skm = f"{skm}://"

                address = f"{skm}{user}{host}{port}{vhost}"
                print(address)
                assert _parse_url(address) == _get_return_tuple(subdict, password=None)

                # special optional tokens
                if user:  # password can only be given alongside username
                    subdict["password"] = "secret"
                    address = f"{skm}{subdict['username']}:{subdict['password']}@{host}{port}{vhost}"
                    print(address)
                    assert _parse_url(address) == _get_return_tuple(subdict)

    def test_200(self) -> None:
        """Test `_get_credentials()`."""

        # Case 1: username/password
        cred = pika.credentials.PlainCredentials("username", "password")
        assert _get_credentials("username", "password", "") == cred
        # Sub-case: username/auth_token (override)
        cred = pika.credentials.PlainCredentials("username", "auth_token")
        assert _get_credentials("username", "password", "auth_token") == cred
        assert _get_credentials("username", None, "auth_token") == cred

        # Error: no password for user
        with pytest.raises(MQClientException):
            _get_credentials("username", None, "")

        # Case 2: Only password/token -- Ex: keycloak
        cred = pika.credentials.PlainCredentials("", "password")
        assert _get_credentials(None, "password", "") == cred
        #
        cred = pika.credentials.PlainCredentials("", "auth_token")
        assert _get_credentials(None, "password", "auth_token") == cred

        # Case 3: no auth -- rabbitmq uses guest/guest
        assert _get_credentials(None, None, "") is None



"""Script for creating a Keycloak realm."""

import asyncio
import json
import os
from functools import partial
from typing import Any

from krs import bootstrap  # type: ignore[import]
from krs.token import get_token  # type: ignore[import]
from rest_tools.client import RestClient


async def keycloak_bootstrap(
    client_id,
    enable_secret=True,
    service_accounts_enabled=False,
    optional_client_scopes=None,
) -> dict:
    """Tools for Keycloack auth integration.

    From https://github.com/WIPACrepo/http-data-transfer-client/blob/main/integration_tests/util.py
    """
    client_secret = bootstrap.bootstrap()

    # monkeypatch.setenv("KEYCLOAK_REALM", "testrealm")  # set env by CI job
    # monkeypatch.setenv("KEYCLOAK_CLIENT_ID", "testclient")  # set env by CI job
    # monkeypatch.setenv("USERNAME", "admin")  # set env by CI job
    # monkeypatch.setenv("PASSWORD", "admin")  # set env by CI job

    # get admin rest client
    token = partial(
        get_token,
        os.environ["KEYCLOAK_URL"],
        client_id=os.environ["KEYCLOAK_CLIENT_ID"],
        client_secret=client_secret,
    )
    rest_client = RestClient(
        f'{os.environ["KEYCLOAK_URL"]}/auth/admin/realms/{os.environ["KEYCLOAK_REALM"]}',
        token=token,
        retries=0,
    )

    # now make http client
    args: Any
    args = {
        "authenticationFlowBindingOverrides": {},
        "bearerOnly": False,
        "clientAuthenticatorType": "client-secret" if enable_secret else "public",
        "clientId": client_id,
        "consentRequired": False,
        "defaultClientScopes": [],
        "directAccessGrantsEnabled": False,
        "enabled": True,
        "frontchannelLogout": False,
        "fullScopeAllowed": True,
        "implicitFlowEnabled": False,
        "notBefore": 0,
        "optionalClientScopes": optional_client_scopes
        if optional_client_scopes
        else [],
        "protocol": "openid-connect",
        "publicClient": False,
        "redirectUris": ["http://localhost*"],
        "serviceAccountsEnabled": service_accounts_enabled,
        "standardFlowEnabled": True,
    }
    await rest_client.request("POST", "/clients", args)

    url = f"/clients?clientId={client_id}"
    ret = await rest_client.request("GET", url)
    if not ret:
        raise Exception("client does not exist")
    data = ret[0]
    keycloak_client_id = data["id"]

    # add mappers
    url = f"/clients/{keycloak_client_id}/protocol-mappers/add-models"
    args = [
        {
            "config": {
                "access.token.claim": "true",
                "access.tokenResponse.claim": "false",
                "claim.name": "authorization_details",
                "claim.value": json.dumps(
                    [
                        {
                            "type": "rabbitmq",
                            "locations": ["cluster:*/vhost:*"],
                            "actions": ["read", "write", "configure"],
                        },
                        {
                            "type": "rabbitmq",
                            "locations": ["cluster:*"],
                            "actions": ["administrator"],
                        },
                    ]
                ),
                "id.token.claim": "false",
                "jsonType.label": "JSON",
                "userinfo.token.claim": "false",
            },
            "consentRequired": False,
            "name": "rich access",
            "protocol": "openid-connect",
            "protocolMapper": "oidc-hardcoded-claim-mapper",
        },
        {
            "config": {
                "access.token.claim": "true",
                "id.token.claim": "false",
                "included.custom.audience": "rabbitmq_client",
            },
            "consentRequired": False,
            "name": "aud-rabbitmq_client",
            "protocol": "openid-connect",
            "protocolMapper": "oidc-audience-mapper",
        },
        {
            "config": {
                "access.token.claim": "true",
                "id.token.claim": "false",
                "included.custom.audience": "rabbitmq",
            },
            "consentRequired": False,
            "name": "aud-rabbitmq",
            "protocol": "openid-connect",
            "protocolMapper": "oidc-audience-mapper",
        },
    ]
    await rest_client.request("POST", url, args)

    # set up return values
    ret_kwargs = {
        "oidc_url": f'{os.environ["KEYCLOAK_URL"]}/auth/realms/{os.environ["KEYCLOAK_REALM"]}',
        "client_id": client_id,
    }
    if enable_secret:
        url = f"/clients/{keycloak_client_id}/client-secret"
        ret = await rest_client.request("GET", url)
        if "value" in ret:
            ret_kwargs["client_secret"] = ret["value"]
        else:
            raise Exception("no client secret")
    return ret_kwargs


async def main() -> None:
    """Do main."""
    kwargs = await keycloak_bootstrap(
        "mqclient-integration-test",
        enable_secret=True,
        service_accounts_enabled=True,
        optional_client_scopes=["profile", "offline_access"],
    )

    # write to files
    with open("KEYCLOAK_OIDC_URL.txt", "w") as f:
        print(kwargs["oidc_url"], file=f)

    with open("KEYCLOAK_CLIENT_ID.txt", "w") as f:
        print(kwargs["client_id"], file=f)

    with open("KEYCLOAK_CLIENT_SECRET.txt", "w") as f:
        print(kwargs["client_secret"], file=f)


if __name__ == "__main__":
    asyncio.run(main())



"""Run integration tests for given broker_client, on broker_client_interface
classes.

Verify functionality that is abstracted away from the Queue class.
"""

import base64
import copy
import itertools
import logging
import pickle
import re
from typing import List, Optional

import asyncstdlib as asl
import pytest

from mqclient.broker_client_interface import BrokerClient, Message
from mqclient.config import DEFAULT_RETRIES, DEFAULT_RETRY_DELAY, DEFAULT_TIMEOUT_MILLIS
from .utils import DATA_LIST, _log_recv, _log_send


def _log_recv_message(recv_msg: Optional[Message]) -> None:
    recv_data = None
    if recv_msg:
        recv_data = recv_msg.data
    _log_recv(f"{recv_msg} -> {recv_data}")


class MyComplexDataForTest01:
    """Used in test 01."""

    def __init__(self, inner_data):
        self.inner_data = inner_data

    def __eq__(self, other) -> bool:
        return self.inner_data == other.inner_data


class PubSubBrokerClientInterface:
    """Integration test suite for broker_client_interface objects.

    Only test things that cannot be tested via the Queue class.
    """

    broker_client: BrokerClient
    timeout = 1

    @pytest.mark.asyncio
    async def test_00(self, queue_name: str, auth_token: str) -> None:
        """Sanity test."""
        pub = await self.broker_client.create_pub_queue(
            "localhost", queue_name, auth_token
        )
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 1, auth_token
        )

        # send
        for msg in DATA_LIST:
            raw_data = Message.serialize(msg)
            await pub.send_message(
                raw_data,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_send(msg)

        # receive
        for i in itertools.count():
            logging.info(i)
            assert i <= len(DATA_LIST)

            recv_msg = await sub.get_message(
                timeout_millis=DEFAULT_TIMEOUT_MILLIS,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_recv_message(recv_msg)

            # check received message
            if i == len(DATA_LIST):
                assert not recv_msg  # None signifies end of stream
                break

            assert recv_msg
            assert DATA_LIST[i] == recv_msg.data

            await sub.ack_message(
                recv_msg,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )

        await pub.close()
        await sub.close()

    @pytest.mark.asyncio
    async def test_01__not_jsonable_data(
        self, queue_name: str, auth_token: str
    ) -> None:
        """Sanity test."""
        pub = await self.broker_client.create_pub_queue(
            "localhost", queue_name, auth_token
        )
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 1, auth_token
        )

        pickable_data_list = [MyComplexDataForTest01(d) for d in DATA_LIST]

        # sanity checks
        with pytest.raises(
            TypeError,
            match=re.escape(
                f"Object of type {type(pickable_data_list[0]).__name__} is not JSON serializable"
            ),
        ):
            Message.serialize(pickable_data_list[0])
        with pytest.raises(
            TypeError,
            match=re.escape("Object of type bytes is not JSON serializable"),
        ):
            Message.serialize(pickle.dumps(pickable_data_list[0]))
        with pytest.raises(
            TypeError,
            match=re.escape("Object of type bytes is not JSON serializable"),
        ):
            Message.serialize(base64.b64encode(pickle.dumps(pickable_data_list[0])))

        # send
        for msg in pickable_data_list:
            raw_data = Message.serialize(
                # obj -> bytes -> bytes -> str
                base64.b64encode(pickle.dumps(msg)).decode("utf-8")
            )
            await pub.send_message(
                raw_data,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_send(msg)

        # receive
        for i in itertools.count():
            logging.info(i)
            assert i <= len(pickable_data_list)

            recv_msg = await sub.get_message(
                timeout_millis=DEFAULT_TIMEOUT_MILLIS,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_recv_message(recv_msg)

            # check received message
            if i == len(pickable_data_list):
                assert not recv_msg  # None signifies end of stream
                break

            assert recv_msg
            # str -> bytes -> obj
            decoded_msg = pickle.loads(base64.b64decode(recv_msg.data))
            assert isinstance(pickable_data_list[i], type(decoded_msg))
            assert pickable_data_list[i] == decoded_msg

            await sub.ack_message(
                recv_msg,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )

        await pub.close()
        await sub.close()

    @pytest.mark.asyncio
    async def test_10(self, queue_name: str, auth_token: str) -> None:
        """Test nacking, front-loaded sending.

        Order is not guaranteed on redelivery.
        """
        pub = await self.broker_client.create_pub_queue(
            "localhost", queue_name, auth_token
        )
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 1, auth_token
        )

        # send
        for msg in DATA_LIST:
            raw_data = Message.serialize(msg)
            await pub.send_message(
                raw_data,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_send(msg)

        # receive -- nack each message, once, and anticipate its redelivery
        nacked_msgs: List[Message] = []
        redelivered_msgs: List[Message] = []
        for i in itertools.count():
            logging.info(i)
            assert i < len(DATA_LIST) * 10  # large enough but avoids inf loop

            # all messages have been acked and redelivered
            if len(redelivered_msgs) == len(DATA_LIST):
                redelivered_data = [m.data for m in redelivered_msgs]
                assert all((d in DATA_LIST) for d in redelivered_data)
                break

            recv_msg = await sub.get_message(
                timeout_millis=DEFAULT_TIMEOUT_MILLIS,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_recv_message(recv_msg)

            if not recv_msg:
                logging.info("waiting...")
                continue
            assert recv_msg.data in DATA_LIST

            # message was redelivered, so ack it
            if recv_msg in nacked_msgs:
                logging.info("REDELIVERED!")
                nacked_msgs.remove(recv_msg)
                redelivered_msgs.append(recv_msg)
                await sub.ack_message(
                    recv_msg,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
            # otherwise, nack message
            else:
                nacked_msgs.append(recv_msg)
                await sub.reject_message(
                    recv_msg,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
                logging.info("NACK!")

        await pub.close()
        await sub.close()

    @pytest.mark.asyncio
    async def test_11(self, queue_name: str, auth_token: str) -> None:
        """Test nacking, mixed sending and receiving.

        Order is not guaranteed on redelivery.
        """
        pub = await self.broker_client.create_pub_queue(
            "localhost", queue_name, auth_token
        )
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 1, auth_token
        )

        data_to_send = copy.deepcopy(DATA_LIST)
        nacked_msgs: List[Message] = []
        redelivered_msgs: List[Message] = []
        for i in itertools.count():
            logging.info(i)
            assert i < len(DATA_LIST) * 10  # large enough but avoids inf loop

            # all messages have been acked and redelivered
            if len(redelivered_msgs) == len(DATA_LIST):
                redelivered_data = [m.data for m in redelivered_msgs]
                assert all((d in DATA_LIST) for d in redelivered_data)
                break

            # send a message
            if data_to_send:
                msg = data_to_send[0]
                raw_data = Message.serialize(msg)
                await pub.send_message(
                    raw_data,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
                _log_send(msg)
                data_to_send.remove(msg)

            # get a message
            recv_msg = await sub.get_message(
                timeout_millis=DEFAULT_TIMEOUT_MILLIS,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_recv_message(recv_msg)

            if not recv_msg:
                logging.info("waiting...")
                continue
            assert recv_msg.data in DATA_LIST

            # message was redelivered, so ack it
            if recv_msg in nacked_msgs:
                logging.info("REDELIVERED!")
                nacked_msgs.remove(recv_msg)
                redelivered_msgs.append(recv_msg)
                await sub.ack_message(
                    recv_msg,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
            # otherwise, nack message
            else:
                nacked_msgs.append(recv_msg)
                await sub.reject_message(
                    recv_msg,
                    retries=DEFAULT_RETRIES,
                    retry_delay=DEFAULT_RETRY_DELAY,
                )
                logging.info("NACK!")

        await pub.close()
        await sub.close()

    @pytest.mark.asyncio
    async def test_20(self, queue_name: str, auth_token: str) -> None:
        """Sanity test message generator."""
        pub = await self.broker_client.create_pub_queue(
            "localhost", queue_name, auth_token
        )
        sub = await self.broker_client.create_sub_queue(
            "localhost", queue_name, 1, auth_token
        )

        # send
        for msg in DATA_LIST:
            raw_data = Message.serialize(msg)
            await pub.send_message(
                raw_data,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
            _log_send(msg)

        # receive
        last = 0
        recv_msg: Optional[Message]
        async for i, recv_msg in asl.enumerate(
            sub.message_generator(
                timeout=self.timeout,
                propagate_error=True,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )
        ):
            logging.info(i)
            _log_recv_message(recv_msg)
            assert recv_msg
            assert recv_msg.data in DATA_LIST
            last = i
            await sub.ack_message(
                recv_msg,
                retries=DEFAULT_RETRIES,
                retry_delay=DEFAULT_RETRY_DELAY,
            )

        assert last == len(DATA_LIST) - 1

        await pub.close()
        await sub.close()



"""Manage the different broker_clients."""

from types import ModuleType
from typing import Dict, Optional

from .broker_client_interface import BrokerClient, MQClientException

# Import all the broker clients at package import, so any bindings can be built/compiled
# fmt: off
_INSTALLED_BROKERS: Dict[str, Optional[ModuleType]] = {}
# Pulsar
try:
    from .broker_clients import apachepulsar
    _INSTALLED_BROKERS["pulsar"] = apachepulsar
except (ModuleNotFoundError, ImportError):
    _INSTALLED_BROKERS["pulsar"] = None
# NATS
try:
    from .broker_clients import nats
    _INSTALLED_BROKERS["nats"] = nats
except (ModuleNotFoundError, ImportError):
    _INSTALLED_BROKERS["nats"] = None
# RabbitMQ
try:
    from .broker_clients import rabbitmq
    _INSTALLED_BROKERS["rabbitmq"] = rabbitmq
except (ModuleNotFoundError, ImportError):
    _INSTALLED_BROKERS["rabbitmq"] = None
# fmt: on


def get_broker_client(broker_client_name: str) -> BrokerClient:
    """Get the `BrokerClient` instance per the given name."""
    try:
        module = _INSTALLED_BROKERS[broker_client_name]
    except KeyError:
        raise MQClientException(f"Unknown broker client: {broker_client_name}")

    if not module:
        raise MQClientException(
            f"Install the '{broker_client_name.lower()}' extra "
            f"if you want to use the '{broker_client_name}' broker client"
        )
    return module.BrokerClient()  # type: ignore[no-any-return]



"""Unit test the broker_client interface."""

# fmt: off

# local imports
from mqclient import broker_client_interface


def test_RawQueue() -> None:
    """Test RawQueue."""
    assert hasattr(broker_client_interface, "RawQueue")


def test_Message() -> None:
    """Test Message."""
    m = broker_client_interface.Message('foo', b'abc')
    assert m.msg_id == 'foo'
    assert m.payload == b'abc'
    assert m._ack_status == broker_client_interface.Message.AckStatus.NONE



"""Back-end using RabbitMQ."""

import copy
import functools
import logging
import urllib
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import pika  # type: ignore

from .. import broker_client_interface, log_msgs
from ..broker_client_interface import (
    ClosingFailedException,
    ConnectingFailedException,
    Message,
    MQClientException,
    Pub,
    RawQueue,
    Sub,
)
from . import utils

StrDict = Dict[str, Any]

LOGGER = logging.getLogger("mqclient.rabbitmq")

HEARTBEAT_STREAMLOSTERROR_MSG = (
    "pika.exceptions.StreamLostError: may be due to a missed heartbeat"
)


HUMAN_PATTERN = "[SCHEME://][USER[:PASS]@]HOST[:PORT][/VIRTUAL_HOST]"


def _parse_url(url: str) -> Tuple[StrDict, Optional[str], Optional[str]]:
    if "://" not in url:
        url = "//" + url
    result = urllib.parse.urlparse(url)

    parts = dict(
        scheme=result.scheme,
        host=result.hostname,
        port=result.port,
        virtual_host=result.path.lstrip("/"),
    )
    # for putting into ConnectionParameters filter ""/None (will rely on defaults)
    parts = {k: v for k, v in parts.items() if v}  # host=..., etc.

    # check validity
    if not parts or "host" not in parts:
        raise MQClientException(f"Invalid address: {url} (format: {HUMAN_PATTERN})")

    return parts, result.username, result.password


def _get_credentials(
    username: Optional[str], password: Optional[str], auth_token: str
) -> Optional[pika.credentials.PlainCredentials]:
    if auth_token:
        password = auth_token

    # Case 1: username/password
    if username and password:
        return pika.credentials.PlainCredentials(username, password)
    # Case 2: Only password/token -- Ex: keycloak
    elif (not username) and password:
        return pika.credentials.PlainCredentials("", password)
    # Error: no password for user
    elif username and (not password):
        raise MQClientException("username given but no password or token")
    # Case 3: no auth -- rabbitmq uses guest/guest
    else:  # not username and not password
        return None


class RabbitMQ(RawQueue):
    """Base RabbitMQ wrapper.

    Extends:
        RawQueue
    """

    def __init__(
        self,
        address: str,
        queue: str,
        auth_token: str,
    ) -> None:
        super().__init__()
        LOGGER.info(f"Requested MQClient for queue '{queue}' @ {address}")
        cp_args, _user, _pass = _parse_url(address)

        # set up connection parameters
        if creds := _get_credentials(_user, _pass, auth_token):
            cp_args["credentials"] = creds

        self.parameters = pika.connection.ConnectionParameters(**cp_args)

        self.queue = queue
        self.connection: Optional[pika.BlockingConnection] = None
        self._next_channel_number = 1  # must start at 1 (not 0)

    def open_channel(self) -> pika.adapters.blocking_connection.BlockingChannel:
        """Open a channel for the connection and configure."""
        LOGGER.info(f"Opening channel to connection for '{self.queue=}'")
        if not self.connection:
            raise ClosingFailedException("No connection to open channel.")

        # give unique channel_number b/c pika has a delay on re-connections in which it will recycle a closed channel
        channel = self.connection.channel(self._next_channel_number)
        # alternatively, we could use: int(uuid.uuid4()) % pika.channel.MAX_CHANNELS + 1
        # the channel number gets put in a struct and it's constrained to an unsigned short
        # 0 is not allowed and will be treated as None
        self._next_channel_number += 1

        """
        We need to discuss how many RabbitMQ instances we want to run
        the default is that the quorum queue is spread across 3 nodes
        so 1 can fail without issue. Maybe we want to up this for
        more production workloads
        """
        channel.queue_declare(
            queue=self.queue, durable=True, arguments={"x-queue-type": "quorum"}
        )

        LOGGER.info(f"Opened channel '{channel.channel_number}'")
        return channel

    async def connect(self) -> pika.adapters.blocking_connection.BlockingChannel:
        """Set up connection and open 1 channel."""
        await super().connect()
        LOGGER.info(f"Connecting with parameters={self.parameters}")

        self.connection = pika.BlockingConnection(self.parameters)
        channel = self.open_channel()
        if not channel:
            raise ConnectingFailedException("Channel was not connected")
        return channel

    async def close(self) -> None:
        """Close connection."""
        await super().close()

        if not self.connection:
            raise ClosingFailedException("No connection to close.")
        if self.connection.is_closed:
            LOGGER.warning("Attempted to close a connection that is already closed")
            return

        try:
            self.connection.close()  # also closes each channel
        except Exception as e:
            raise ClosingFailedException() from e


class RabbitMQPub(RabbitMQ, Pub):
    """Wrapper around queue with delivery-confirm mode in the channel.

    Extends:
        RabbitMQ
        Pub
    """

    def __init__(
        self,
        address: str,
        name: str,
        auth_token: str,
    ) -> None:
        LOGGER.debug(f"{log_msgs.INIT_PUB} ({address}; {name})")
        super().__init__(address, name, auth_token)
        self.channel: Optional[pika.adapters.blocking_connection.BlockingChannel] = None

    def open_channel(self) -> pika.adapters.blocking_connection.BlockingChannel:
        """Open a channel for the connection and configure."""
        if self.channel:
            raise MQClientException("RabbitMQPub instance can only have one channel")

        channel = super().open_channel()
        channel.confirm_delivery()
        return channel

    async def connect(self) -> None:
        """Set up connection, channel, and queue.

        Turn on delivery confirmations.
        """
        LOGGER.debug(log_msgs.CONNECTING_PUB)
        self.channel = await super().connect()
        LOGGER.debug(log_msgs.CONNECTED_PUB)

    async def close(self) -> None:
        """Close connection."""
        LOGGER.debug(log_msgs.CLOSING_PUB)
        await super().close()
        self.channel = None
        LOGGER.debug(log_msgs.CLOSED_PUB)

    async def send_message(
        self,
        msg: bytes,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Send a message on a queue.

        Args:
            address (str): address of queue
            name (str): name of queue on address

        Returns:
            RawQueue: queue
        """
        LOGGER.debug(log_msgs.SENDING_MESSAGE)
        if not self.channel:
            raise MQClientException("queue is not connected")

        def _send_msg():
            # use wrapper function so connection references can be updated by reconnects
            if not self.channel:
                raise MQClientException("queue is not connected")
            LOGGER.debug(f"sending on channel: {self.channel.channel_number}")
            return self.channel.basic_publish(
                exchange="",
                routing_key=self.queue,
                body=msg,
            )

        try:
            await utils.auto_retry_call(
                func=_send_msg,
                nonretriable_conditions=lambda e: isinstance(
                    e,
                    (pika.exceptions.AMQPChannelError, pika.exceptions.StreamLostError),
                ),
                retries=retries,
                retry_delay=retry_delay,
                close=self.close,
                connect=self.connect,
                logger=LOGGER,
            )
            LOGGER.debug(log_msgs.SENT_MESSAGE)
        except pika.exceptions.StreamLostError as e:
            raise MQClientException(HEARTBEAT_STREAMLOSTERROR_MSG) from e


class RabbitMQSub(RabbitMQ, Sub):
    """Wrapper around queue with prefetch-queue QoS.

    Extends:
        RabbitMQ
        Sub
    """

    def __init__(
        self,
        address: str,
        name: str,
        auth_token: str,
        prefetch: int,
    ) -> None:
        LOGGER.debug(f"{log_msgs.INIT_SUB} ({address}; {name})")
        super().__init__(address, name, auth_token)
        self.consumer_id = None
        self.prefetch = prefetch

        self.active_channels: List[
            pika.adapters.blocking_connection.BlockingChannel
        ] = []
        self.reserve_channel: Optional[
            pika.adapters.blocking_connection.BlockingChannel
        ] = None

    def open_channel(self) -> pika.adapters.blocking_connection.BlockingChannel:
        """Open a channel for the connection and configure.

        Turn on prefetching.
        """
        channel = super().open_channel()

        channel.basic_qos(prefetch_count=max(self.prefetch, 1))
        # Setting the value to 0 lets the consumer drain the entire queue.
        # https://www.cloudamqp.com/blog/rabbitmq-basic-consume-vs-rabbitmq-basic-get.html#what-are-the-advantages-of-a-rabbitmq-consumer
        # https://www.cloudamqp.com/blog/part1-rabbitmq-best-practice.html#prefetch
        #
        # https://www.rabbitmq.com/consumer-prefetch.html
        #    Meaning of prefetch_count in RabbitMQ w/ global_qos=False:
        #    applied separately to each new consumer on the channel
        #
        # global_qos=False b/c using quorum queues
        # https://www.rabbitmq.com/quorum-queues.html#global-qos

        return channel

    async def connect(self) -> None:
        """Set up connection, channel, and queue."""
        LOGGER.debug(log_msgs.CONNECTING_SUB)
        channel = await super().connect()
        self.active_channels.append(channel)
        LOGGER.debug(log_msgs.CONNECTED_SUB)

    async def close(self) -> None:
        """Close connection.

        Also, channel will be canceled (rejects all pending ackable
        messages).
        """
        LOGGER.debug(log_msgs.CLOSING_SUB)
        await super().close()
        for channel in self.active_channels:
            if channel.is_open:
                LOGGER.warning(
                    f"Channel remains open after connection close: {channel.channel_number}."
                )
        if self.reserve_channel and self.reserve_channel.is_open:
            LOGGER.warning(
                f"Reserve channel remains open after connection close: {self.reserve_channel.channel_number}."
            )
        self.active_channels = []
        self.reserve_channel = None
        LOGGER.debug(log_msgs.CLOSED_SUB)

    @staticmethod
    def _to_message(  # noqa: F821 # pylint: disable=W0221
        method_frame: Optional[pika.spec.Basic.GetOk],
        body: Optional[Union[str, bytes]],
        channel_number: int,
    ) -> Optional[Message]:
        """Transform RabbitMQ-Message to Message type."""
        if not method_frame or body is None:
            return None

        if isinstance(body, str):
            msg = Message(method_frame.delivery_tag, body.encode())
        else:
            msg = Message(method_frame.delivery_tag, body)

        msg._connection_id = channel_number
        return msg

    async def _iter_messages(
        self,
        timeout_millis: Optional[int],
        retries: int,
        retry_delay: float,
    ) -> AsyncIterator[Optional[Message]]:
        if not self.active_channels:
            raise MQClientException("queue is not connected")

        def _get_msg():
            # use wrapper function so connection references can be updated by reconnects
            if not self.active_channels:
                raise MQClientException("queue is not connected")
            LOGGER.debug(f"consuming on channel: {channel.channel_number}")
            try:
                return next(
                    # pika smartly handles re-invocations
                    channel.consume(
                        self.queue,
                        inactivity_timeout=timeout_millis / 1000.0
                        if timeout_millis
                        else None,
                    )
                )
            except StopIteration:
                return (None, None, None)

        remaining_active_channels = copy.copy(self.active_channels)  # start with all
        channel = remaining_active_channels[0]

        while True:
            try:
                pika_msg = await utils.auto_retry_call(
                    func=_get_msg,
                    nonretriable_conditions=lambda e: isinstance(
                        e,
                        (
                            pika.exceptions.AMQPChannelError,
                            pika.exceptions.StreamLostError,
                        ),
                    ),
                    retries=retries,
                    retry_delay=retry_delay,
                    close=None,
                    connect=None,
                    logger=LOGGER,
                )
            except pika.exceptions.StreamLostError as e:
                raise MQClientException(HEARTBEAT_STREAMLOSTERROR_MSG) from e

            #
            # YIELD (got a message)
            if msg := RabbitMQSub._to_message(
                pika_msg[0],
                pika_msg[2],
                channel.channel_number,
            ):
                LOGGER.debug(f"{log_msgs.GETMSG_RECEIVED_MESSAGE} ({msg.msg_id!r}).")
                if channel == self.reserve_channel:
                    # if this was the reserve channel, move it to active channels
                    self.active_channels.append(self.reserve_channel)
                    self.reserve_channel = None  # no need to open a new one now
                remaining_active_channels = copy.copy(self.active_channels)  # reset!
                yield msg
            #
            # DEAL WITH EMPTY CHANNEL (didn't get a message)
            else:
                if channel == self.reserve_channel:
                    LOGGER.debug("No message received -- tried all channels")
                    # this means our reserve channel came up empty,
                    # so there's REALLY nothing in the queue
                    LOGGER.debug(log_msgs.GETMSG_NO_MESSAGE)
                    remaining_active_channels = copy.copy(self.active_channels)  # reset
                    yield None
                else:
                    LOGGER.debug("No message received -- switching channels...")
                    remaining_active_channels.remove(channel)
                    if not remaining_active_channels:
                        # try reserve channel
                        if not self.reserve_channel:
                            self.reserve_channel = self.open_channel()
                        channel = self.reserve_channel
                    else:
                        # try next active channel
                        channel = remaining_active_channels[0]

    async def get_message(
        self,
        timeout_millis: Optional[int],
        retries: int,
        retry_delay: float,
    ) -> Optional[Message]:
        """Get a message from a queue."""
        LOGGER.debug(log_msgs.GETMSG_RECEIVE_MESSAGE)
        if not self.active_channels:
            raise MQClientException("queue is not connected")

        msg = None
        async for msg in self._iter_messages(timeout_millis, retries, retry_delay):
            # None -> timeout
            break  # get just one message

        return msg

    def _get_channel_by_msg(
        self, msg: Message
    ) -> pika.adapters.blocking_connection.BlockingChannel:
        """Map message to channel."""
        matches = [
            c for c in self.active_channels if c.channel_number == msg._connection_id
        ]
        if not matches:
            raise MQClientException(
                f"could not map message to channel: {msg} {self.active_channels}"
            )
        elif len(matches) > 1:
            raise MQClientException(
                f"message mapped to multiple channels: {msg} {matches}"
            )
        else:
            return matches[0]

    async def ack_message(
        self,
        msg: Message,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Ack a message from the queue."""
        LOGGER.debug(log_msgs.ACKING_MESSAGE)
        if not self.active_channels:
            raise MQClientException("queue is not connected")

        channel = self._get_channel_by_msg(msg)
        LOGGER.debug(f"acking on channel: {channel.channel_number}")

        try:
            await utils.auto_retry_call(
                func=functools.partial(
                    channel.basic_ack,
                    msg.msg_id,
                    multiple=False,
                ),
                connect=None,
                close=None,
                nonretriable_conditions=lambda e: isinstance(
                    e,
                    (pika.exceptions.AMQPChannelError, pika.exceptions.StreamLostError),
                ),
                retries=retries,
                retry_delay=retry_delay,
                logger=LOGGER,
            )
            LOGGER.debug(f"{log_msgs.ACKED_MESSAGE} ({msg.msg_id!r}).")
        except pika.exceptions.StreamLostError as e:
            raise MQClientException(HEARTBEAT_STREAMLOSTERROR_MSG) from e

    async def reject_message(
        self,
        msg: Message,
        retries: int,
        retry_delay: float,
    ) -> None:
        """Reject (nack) a message from the queue."""
        LOGGER.debug(log_msgs.NACKING_MESSAGE)
        if not self.active_channels:
            raise MQClientException("queue is not connected")

        channel = self._get_channel_by_msg(msg)
        LOGGER.debug(f"nacking on channel: {channel.channel_number}")

        try:
            await utils.auto_retry_call(
                func=functools.partial(
                    channel.basic_nack,
                    msg.msg_id,
                    multiple=False,
                    requeue=True,
                ),
                close=None,
                connect=None,
                nonretriable_conditions=lambda e: isinstance(
                    e,
                    (pika.exceptions.AMQPChannelError, pika.exceptions.StreamLostError),
                ),
                retries=retries,
                retry_delay=retry_delay,
                logger=LOGGER,
            )
            LOGGER.debug(f"{log_msgs.NACKED_MESSAGE} ({msg.msg_id!r}).")
        except pika.exceptions.StreamLostError as e:
            raise MQClientException(HEARTBEAT_STREAMLOSTERROR_MSG) from e

    async def message_generator(
        self,
        timeout: int,
        propagate_error: bool,
        retries: int,
        retry_delay: float,
    ) -> AsyncGenerator[Optional[Message], None]:
        """Yield Messages.

        Generate messages with variable timeout.
        Yield `None` on `throw()`.

        Keyword Arguments:
            timeout {int} -- timeout in seconds for inactivity (default: {60})
            propagate_error -- should errors from downstream kill the generator? (default: {True})
        """
        LOGGER.debug(log_msgs.MSGGEN_ENTERED)
        if not self.active_channels:
            raise MQClientException("queue is not connected")

        msg = None
        try:
            async for msg in self._iter_messages(timeout * 1000, retries, retry_delay):
                LOGGER.debug(log_msgs.MSGGEN_GET_NEW_MESSAGE)
                if not msg:
                    LOGGER.info(log_msgs.MSGGEN_NO_MESSAGE_LOOK_BACK_IN_QUEUE)
                    break

                # yield message to consumer
                try:
                    LOGGER.debug(f"{log_msgs.MSGGEN_YIELDING_MESSAGE} [{msg}]")
                    yield msg
                # consumer throws Exception...
                except Exception as e:  # pylint: disable=W0703
                    LOGGER.debug(log_msgs.MSGGEN_DOWNSTREAM_ERROR)
                    if propagate_error:
                        LOGGER.debug(log_msgs.MSGGEN_PROPAGATING_ERROR)
                        raise
                    LOGGER.warning(
                        f"{log_msgs.MSGGEN_EXCEPTED_DOWNSTREAM_ERROR} {e}.",
                        exc_info=True,
                    )
                    yield None  # hand back to consumer
                # consumer requests again, aka next()
                else:
                    pass

        # Garbage Collection (or explicit generator close(), or break in consumer's loop)
        except GeneratorExit:
            LOGGER.debug(log_msgs.MSGGEN_GENERATOR_EXITING)
            LOGGER.debug(log_msgs.MSGGEN_GENERATOR_EXITED)

        # Done with generator, one way or another
        finally:
            pass


class BrokerClient(broker_client_interface.BrokerClient):
    """RabbitMQ Pub-Sub BrokerClient Factory.

    Extends:
        BrokerClient
    """

    NAME = "rabbitmq"

    @staticmethod
    async def create_pub_queue(
        address: str,
        name: str,
        auth_token: str,
    ) -> RabbitMQPub:
        """Create a publishing queue.

        Args:
            address (str): address of queue
            name (str): name of queue on address

        Returns:
            RawQueue: queue
        """
        # pylint: disable=invalid-name
        q = RabbitMQPub(address, name, auth_token)
        await q.connect()
        return q

    @staticmethod
    async def create_sub_queue(
        address: str,
        name: str,
        prefetch: int,
        auth_token: str,
    ) -> RabbitMQSub:
        """Create a subscription queue.

        Args:
            address (str): address of queue
            name (str): name of queue on address

        Returns:
            RawQueue: queue
        """
        # pylint: disable=invalid-name
        q = RabbitMQSub(address, name, auth_token, prefetch)
        await q.connect()
        return q



"""Queue class encapsulating a pub-sub messaging system."""

import contextlib
import logging
import os
import sys
import types
import uuid
from typing import Any, AsyncGenerator, AsyncIterator, Dict, Optional, Type

from . import broker_client_manager
from . import telemetry as wtt
from .broker_client_interface import (
    AckException,
    Message,
    MQClientException,
    NackException,
    Pub,
    Sub,
)
from .config import (
    DEFAULT_EXCEPT_ERRORS,
    DEFAULT_PREFETCH,
    DEFAULT_RETRIES,
    DEFAULT_RETRY_DELAY,
    DEFAULT_TIMEOUT,
    MIN_PREFETCH,
)

LOGGER = logging.getLogger("mqclient")


# deprecation check
for envvar in ["RABBITMQ_HEARTBEAT"]:
    if os.getenv(envvar):
        raise MQClientException(f"Environment variable {envvar} has been deprecated.")


def _message_size_message(msg: Message) -> str:
    return (
        f"{sys.getsizeof(msg.payload)} bytes "
        f"(data={sys.getsizeof(msg.data)}, headers={sys.getsizeof(msg.headers)}) "
        f"[msg_id={msg.msg_id!r}]"
    )


class Queue:
    """User-facing queue library.

    Args:
        broker_client: the broker_client to use
        address: address of queue
        name: name of queue
        prefetch: size of prefetch buffer for receiving messages (min 1)
        timeout: seconds to wait for a message to be delivered
        except_errors: whether to suppress interior context errors for
                        the consumer (when `True`, the context manager
                        will act like a `try-except` block)
        auth_token: the (jwt) authentication token
    """

    def __init__(
        self,
        broker_client: str,
        address: str = "localhost",
        name: str = "",
        prefetch: int = DEFAULT_PREFETCH,
        timeout: int = DEFAULT_TIMEOUT,
        retry_delay: float = DEFAULT_RETRY_DELAY,  # seconds
        retries: int = DEFAULT_RETRIES,  # ex: 2 means 1 initial try and 2 retries
        except_errors: bool = DEFAULT_EXCEPT_ERRORS,
        auth_token: str = "",
    ) -> None:
        self._broker_client = broker_client_manager.get_broker_client(broker_client)
        self._address = address
        self._name = name if name else Queue.make_name()

        if prefetch < MIN_PREFETCH:
            raise ValueError(f"prefetch must be >= {MIN_PREFETCH}")
        self._prefetch = prefetch

        self._auth_token = auth_token

        # properties
        self._timeout = -1
        self.timeout = timeout
        self._retries = -1
        self.retries = retries
        self._retry_delay = -1.0
        self.retry_delay = retry_delay

        # publics
        self.except_errors = except_errors

    @staticmethod
    def make_name() -> str:
        """Return a pseudo-unique string that is a legal queue identifier.

        This name is valid for any broker_client chosen.
        """
        return "a" + (uuid.uuid4().hex)[:20]

    @property
    def timeout(self) -> int:
        """Get the timeout value."""
        return self._timeout

    @timeout.setter
    def timeout(self, val: int) -> None:
        LOGGER.debug(f"Setting timeout to {val}")
        if val < 1:
            raise ValueError("timeout must be positive")
        self._timeout = val

    @property
    def retries(self) -> int:
        """Get the retries value."""
        return self._retries

    @retries.setter
    def retries(self, val: int) -> None:
        LOGGER.debug(f"Setting retries to {val}")
        if val < 0:
            raise ValueError("retries must be non-negative")
        self._retries = val

    @property
    def retry_delay(self) -> float:
        """Get the retry_delay value."""
        return self._retry_delay

    @retry_delay.setter
    def retry_delay(self, val: float) -> None:
        LOGGER.debug(f"Setting retry_delay to {val}")
        if val <= 0:
            raise ValueError("retry_delay must be positive")
        self._retry_delay = val

    async def _create_pub_queue(self) -> Pub:
        """Wrap `self._broker_client.create_pub_queue()` with instance's
        config."""
        return await self._broker_client.create_pub_queue(
            self._address,
            self._name,
            self._auth_token,
        )

    async def _create_sub_queue(self, prefetch_override: Optional[int] = None) -> Sub:
        """Wrap `self._broker_client.create_sub_queue()` with instance's
        config."""
        return await self._broker_client.create_sub_queue(
            self._address,
            self._name,
            # 0 is okay
            prefetch_override if prefetch_override is not None else self._prefetch,
            self._auth_token,
        )

    @contextlib.asynccontextmanager  # needs to wrap @wtt stuff to span children correctly
    @wtt.spanned(
        these=[
            "self._broker_client",
            "self._address",
            "self._name",
            "self._prefetch",
            "self.timeout",
        ]
    )
    async def open_pub(self) -> AsyncIterator["QueuePubResource"]:
        """Open a resource to send messages to the queue.

        This is an async context manager. An object is returned that can be
        used to send n messages.

        Example:
            async with queue.open_pub() as p:
                for msg in my_messages:
                    await p.send(msg)

        Decorators:
            contextlib.asynccontextmanager

        Returns:
            QueuePubResource -- the object to invoke `.send()` on
        """
        pub = await self._create_pub_queue()

        try:
            yield QueuePubResource(pub, self.retries, self.retry_delay)
        finally:
            await pub.close()

    @wtt.spanned(
        these=[
            "self._broker_client",
            "self._address",
            "self._name",
            "self._prefetch",
            "self.timeout",
            "msg.msg_id",
        ]
    )
    async def _safe_ack(self, sub: Sub, msg: Message) -> None:
        """Acknowledge the message."""
        # pylint:disable=protected-access
        if msg._ack_status == Message.AckStatus.NONE:
            try:
                await sub.ack_message(
                    msg,
                    retries=self.retries,
                    retry_delay=self.retry_delay,
                )
                msg._ack_status = Message.AckStatus.ACKED  # mark after success
            except Exception as e:
                raise AckException(f"Acking failed on broker_client: {msg}") from e
        elif msg._ack_status == Message.AckStatus.NACKED:
            raise AckException(
                f"Message has already been nacked, it cannot be acked: {msg}"
            )
        elif msg._ack_status == Message.AckStatus.ACKED:
            # needless, so we'll skip it
            LOGGER.debug(f"Attempted to ack an already-acked message: {msg}")
        else:
            raise MQClientException(f"Unrecognized AckStatus value: {msg}")

    @wtt.spanned(
        these=[
            "self._broker_client",
            "self._address",
            "self._name",
            "self._prefetch",
            "self.timeout",
            "msg.msg_id",
        ]
    )
    async def _safe_nack(self, sub: Sub, msg: Message) -> None:
        """Reject/nack the message."""
        # pylint:disable=protected-access
        if msg._ack_status == Message.AckStatus.NONE:
            try:
                await sub.reject_message(
                    msg,
                    retries=self.retries,
                    retry_delay=self.retry_delay,
                )
                msg._ack_status = Message.AckStatus.NACKED  # mark after success
            except Exception as e:
                raise NackException(f"Nacking failed on broker_client: {msg}") from e
        elif msg._ack_status == Message.AckStatus.NACKED:
            # needless, so we'll skip it
            LOGGER.debug(f"Attempted to nack an already-nacked message: {msg}")
        elif msg._ack_status == Message.AckStatus.ACKED:
            raise NackException(
                f"Message has already been acked, it cannot be nacked: {msg}"
            )
        else:
            raise MQClientException(f"Unrecognized AckStatus value: {msg}")

    def open_sub(self) -> "QueueSubResource":
        """Open a resource to receive messages from the queue as an iterator.

        This returns a context-manager/generator. Its iterator stops when no
        messages are received for `timeout` seconds. If an exception is raised
        (inside the context), the message is rejected, the context is exited,
        and exception is re-raised if configured by `except_errors`. The
        connection is closed after the context manager exits.

        Multiple calls to `open_sub()` is okay, but reusing the returned
        instance is not.

        Example:
            async with queue.open_sub() as sub:
                async for msg in sub:
                    print(msg)

        Returns:
            QueueSubResource -- context manager and generator object
        """
        LOGGER.debug("Creating new QueueSubResource instance.")
        return QueueSubResource(self)

    @contextlib.asynccontextmanager  # needs to wrap @wtt stuff to span children correctly
    @wtt.spanned(
        these=[
            "self._broker_client",
            "self._address",
            "self._name",
            "self._prefetch",
            "self.timeout",
        ]
    )
    async def open_sub_manual_acking(
        self,
    ) -> AsyncGenerator["ManualQueueSubResource", None]:
        """Open a resource to receive messages from the queue as an iterator.

        This returns a context-manager with an iterator function `iter_messages()`.
        The iterator stops when no messages are received for `timeout` seconds.
        Multiple calls to `open_sub()` is okay, but reusing the returned
        instance is not.

        The connection is closed after the context manager exits.

        *Unlike `open_sub()`*, the caller is responsible for:
            - All acking and/or nacking
            - Any error handling

        **NOTE: unless you need to parallelize your message processing,
        use `open_sub()`**

        Examples:
            async with queue.open_sub_manual_acking() as sub:
                async for msg in sub.iter_messages():
                    print(msg.data)
                    sub.ack(msg)
                    # if you choose not to nack on an error,
                    # the broker will take longer to redeliver

            async with queue.open_sub_manual_acking() as sub:
                async for msg in sub.iter_messages():
                    try:
                        process_message(msg.data)
                    except Exception:
                        await sub.nack(msg)
                    else:
                        await sub.ack(msg)

        Returns:
            ManualQueueSubResource -- context manager w/ iterator function
        """
        sub = await self._create_sub_queue()
        resource = ManualQueueSubResource(self, sub)
        try:
            yield resource
        finally:
            await sub.close()

    @contextlib.asynccontextmanager  # needs to wrap @wtt stuff to span children correctly
    @wtt.spanned(
        these=[
            "self._broker_client",
            "self._address",
            "self._name",
            "self._prefetch",
            "self.timeout",
        ]
    )
    async def open_sub_one(self) -> AsyncIterator[Any]:
        """Open a context to receive a single messages from the queue.

        This is an async context manager. If an exception is raised
        (inside the context), the message is rejected, the context is
        exited, and exception can be re-raised if configured by
        `except_errors`.

        Example:
            async with q.open_sub_one() as msg:
                print(msg)

        Decorators:
            contextlib.asynccontextmanager

        Raises:
            EmptyQueueException -- if there is no available message

        Yields:
            Any -- object of data received
        """

        @wtt.spanned(
            kind=wtt.SpanKind.CONSUMER,
            carrier="msg.headers",
            carrier_relation=wtt.CarrierRelation.LINK,
        )
        def add_span_link(msg: Message) -> Message:
            return msg

        sub = await self._create_sub_queue()
        raw_msg = await sub.get_message(
            self.timeout * 1000,
            retries=self.retries,
            retry_delay=self.retry_delay,
        )

        if not raw_msg:  # no message -> close and exit
            await sub.close()
            raise EmptyQueueException(
                "No message is available (`timeout` value may be too low)"
            )

        msg = add_span_link(raw_msg)  # got a message -> link and proceed
        LOGGER.info(f"Received Message: {_message_size_message(msg)}")

        try:
            yield msg.data
        except Exception:  # pylint:disable=broad-except
            await self._safe_nack(sub, msg)
            if not self.except_errors:
                raise
        else:
            await self._safe_ack(sub, msg)
        finally:
            await sub.close()

    def __repr__(self) -> str:
        """Return string of basic properties/attributes."""
        return (
            f"Queue("
            f"{self._broker_client.__module__}, "
            f"address={self._address}, "
            f"name={self._name}, "
            f"prefetch={self._prefetch}, "
            f"timeout={self.timeout}"
            f")"
        )


class EmptyQueueException(Exception):
    """Raised when the queue is empty."""


class QueuePubResource:
    """A manager class around `Pub.send_message()`."""

    def __init__(self, pub: Pub, retries: int, retry_delay: float):
        self.pub = pub
        self.retries = retries
        self.retry_delay = retry_delay

    @wtt.spanned(kind=wtt.SpanKind.PRODUCER)
    async def send(self, data: Any) -> None:
        """Send a message.

        Data must be JSON serializable.
        """
        msg_bytes = Message.serialize(data, headers=wtt.inject_links_carrier())
        LOGGER.info(f"Sending Message: {sys.getsizeof(msg_bytes)} bytes")
        await self.pub.send_message(
            msg_bytes,
            retries=self.retries,
            retry_delay=self.retry_delay,
        )


class ManualQueueSubResource:
    """A manager class around `Sub.get_message()`."""

    def __init__(self, queue: Queue, sub: Sub) -> None:
        self.queue = queue
        self._sub: Sub = sub

    async def iter_messages(self) -> AsyncIterator[Message]:
        """Yield a message."""

        @wtt.spanned(
            kind=wtt.SpanKind.CONSUMER,
            carrier="msg.headers",
            carrier_relation=wtt.CarrierRelation.LINK,
        )
        def add_span_link(msg: Message) -> Message:
            return msg

        while True:
            if not (raw_msg := await self._get(self._sub)):
                LOGGER.debug("sub had no message")
                return
            msg = add_span_link(raw_msg)  # got a message -> link and proceed
            LOGGER.info(f"Received Message: {_message_size_message(msg)}")
            yield msg

    async def _get(self, sub: Sub) -> Optional[Message]:
        return await sub.get_message(
            self.queue.timeout * 1000,
            retries=self.queue.retries,
            retry_delay=self.queue.retry_delay,
        )

    async def ack(self, msg: Message) -> None:
        """Acknowledge the message."""
        await self.queue._safe_ack(self._sub, msg)

    async def nack(self, msg: Message) -> None:
        """Reject/nack the message."""
        await self.queue._safe_nack(self._sub, msg)


class QueueSubResource:
    """An async context-manager generator, wraps `Sub.message_generator()`."""

    RUNTIME_ERROR_CONTEXT_STRING = (
        "'QueueSubResource' object's runtime "
        "context has not been entered. Use 'async with ... as ...' syntax."
    )

    def __init__(self, queue: Queue) -> None:
        LOGGER.debug("[QueueSubResource.__init__()]")
        self.queue = queue

        self._sub: Optional[Sub] = None
        self._gen: Optional[AsyncGenerator[Optional[Message], None]] = None

        self._span: Optional[wtt.Span] = None
        self._span_carrier: Optional[Dict[str, Any]] = None

        self.msg: Optional[Message] = None

    @wtt.spanned(
        these=[
            "self.queue._broker_client",
            "self.queue._address",
            "self.queue._name",
            "self.queue._prefetch",
            "self.queue.timeout",
        ],
        behavior=wtt.SpanBehavior.ONLY_END_ON_EXCEPTION,
    )
    async def __aenter__(self) -> "QueueSubResource":
        """Return instance.

        Triggered by 'with ... as'.
        """
        LOGGER.debug("[QueueSubResource.__aenter__()] entered `with-as` block")

        if self._sub and self._gen:
            raise MQClientException(
                "A 'QueueSubResource' instance cannot be re-entered."
            )

        self._sub = await self.queue._create_sub_queue()
        self._gen = self._sub.message_generator(
            timeout=self.queue.timeout,
            propagate_error=(not self.queue.except_errors),
            retries=self.queue.retries,
            retry_delay=self.queue.retry_delay,
        )

        self._span = wtt.get_current_span()
        self._span_carrier = wtt.inject_span_carrier()

        return self

    @wtt.respanned(
        "self._span",
        behavior=wtt.SpanBehavior.END_ON_EXIT,  # end what was opened by `__aenter__()`
    )
    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[types.TracebackType],
    ) -> bool:
        """Return `True` to suppress any Exception raised by consumer code.

        Return `False` to re-raise/propagate that Exception.

        Arguments:
            exc_type {Optional[BaseException]} -- Exception type.
            exc_val {Optional[Type[BaseException]]} -- Exception object.
            exc_tb {Optional[types.TracebackType]} -- Exception Traceback.
        """
        LOGGER.debug(
            f"[QueueSubResource.__aexit__()] exiting `with-as` block (exc:{exc_type})"
        )
        if not (self._sub and self._gen):
            raise MQClientException(self.RUNTIME_ERROR_CONTEXT_STRING)

        reraise_exception = False

        # Exception Was Raised
        if exc_type and exc_val:
            if self.msg:
                await self.queue._safe_nack(self._sub, self.msg)
            # see how the generator wants to handle the exception
            try:
                # `athrow` is caught by the generator's try-except around `yield`
                await self._gen.athrow(exc_type, exc_val, exc_tb)
            except exc_type:  # message_generator re-raised Exception
                reraise_exception = True
        # Good Exit (No Original Exception)
        else:
            # ack if there was a message yielded (unless it was already nacked)
            if self.msg and self.msg._ack_status != Message.AckStatus.NACKED:
                await self.queue._safe_ack(self._sub, self.msg)

        await self._sub.close()  # close after cleanup

        if reraise_exception:
            LOGGER.debug("[QueueSubResource.__aexit__()] exited & propagated error.")
            return False  # propagate the Exception!
        else:
            # either no exception or suppress the exception
            if exc_type and exc_val:
                LOGGER.debug(
                    "[QueueSubResource.__aexit__()] exited & suppressed error."
                )
            else:
                LOGGER.debug("[QueueSubResource.__aexit__()] exited w/o error.")
            return True  # suppress any Exception

    def __aiter__(self) -> "QueueSubResource":
        """Return instance.

        Triggered with 'for'/'aiter()'.
        """
        LOGGER.debug("[QueueSubResource.__aiter__()] entered loop/`aiter()`")
        if not (self._sub and self._gen):
            raise MQClientException(self.RUNTIME_ERROR_CONTEXT_STRING)
        return self

    @wtt.spanned(
        these=[
            "self.queue._broker_client",
            "self.queue._address",
            "self.queue._name",
            "self.queue._prefetch",
            "self.queue.timeout",
        ],
        carrier="self._span_carrier",
    )
    async def __anext__(self) -> Any:
        """Return next Message in queue."""
        LOGGER.debug("[QueueSubResource.__anext__()] next iteration...")
        if not (self._sub and self._gen):
            raise MQClientException(self.RUNTIME_ERROR_CONTEXT_STRING)

        # ack the previous message before getting a new one (unless it was already nacked)
        if self.msg and self.msg._ack_status != Message.AckStatus.NACKED:
            await self.queue._safe_ack(self._sub, self.msg)

        @wtt.spanned(
            kind=wtt.SpanKind.CONSUMER,
            carrier="msg.headers",
            carrier_relation=wtt.CarrierRelation.LINK,
        )
        def get_message_callback(msg: Optional[Message]) -> Optional[Message]:
            return msg

        try:
            self.msg = get_message_callback(await self._gen.__anext__())
        except StopAsyncIteration:
            self.msg = None  # signal there is no message to ack/nack in `__aexit__()`
            LOGGER.debug(
                "[QueueSubResource.__anext__()] end of loop (StopAsyncIteration)"
            )
            raise

        if not self.msg:
            raise MQClientException(
                "Yielded value is `None`. This should not have happened."
            )

        LOGGER.info(f"Received Message: {_message_size_message(self.msg)}")
        return self.msg.data

    @wtt.spanned(
        these=[
            "self.queue._broker_client",
            "self.queue._address",
            "self.queue._name",
            "self.queue._prefetch",
            "self.queue.timeout",
        ],
    )
    async def nack_current(self) -> None:
        """Manually nack the current (most recently yielded) message."""
        if not (self._sub and self._gen):
            raise MQClientException(self.RUNTIME_ERROR_CONTEXT_STRING)
        if not self.msg:  # case: calling after iterator stopped (unusual but possible)
            return
        # pylint:disable=protected-access
        await self.queue._safe_nack(self._sub, self.msg)



"""Common logging strings."""

INIT_PUB = "[__init__()] Initializing pub."
INIT_SUB = "[__init__()] Initializing sub."

CONNECTING_PUB = "[connect()] Pub connecting..."
CONNECTED_PUB = "[connect()] Pub connected."
CONNECTING_SUB = "[connect()] Sub connecting..."
CONNECTED_SUB = "[connect()] Sub connected."

SENDING_MESSAGE = "[send_message()] Sending message..."
SENT_MESSAGE = "[send_message()] Sent message."

CLOSED_PUB = "[close()] Pub closed."
CLOSING_PUB = "[close()] Closing pub..."
CLOSED_SUB = "[close()] Sub closed."
CLOSING_SUB = "[close()] Closing sub..."

GETMSG_RECEIVE_MESSAGE = "[get_message()] Trying to receive message..."
GETMSG_RECEIVED_MESSAGE = "[get_message()] Received message."
GETMSG_NO_MESSAGE = "[get_message()] Didn't receive message. Returning None."
GETMSG_TIMEOUT_ERROR = "[get_message()] Timeout error. Returning None."
GETMSG_CONNECTION_ERROR_TRY_AGAIN = "[get_message()] Connection error. Trying again."
GETMSG_RAISE_OTHER_ERROR = "[get_message()] Other error. Raising Exception."
GETMSG_CONNECTION_ERROR_MAX_RETRIES = (
    "[get_message()] Connection error. Reached max retries. Raising Exception."
)

ACKING_MESSAGE = "[ack_message()] Ack'ing message..."
ACKED_MESSAGE = "[ack_message()] Ack'd message."

NACKING_MESSAGE = "[reject_message()] Nack'ing message..."
NACKED_MESSAGE = "[reject_message()] Nack'd message."

MSGGEN_ENTERED = "[message_generator()] Entered generator."
MSGGEN_GET_NEW_MESSAGE = "[message_generator()] Getting a new message..."
MSGGEN_NO_MESSAGE_LOOK_BACK_IN_QUEUE = (
    "[message_generator()] No messages in idle timeout window."
)
MSGGEN_YIELDING_MESSAGE = "[message_generator()] Yielding message..."
MSGGEN_DOWNSTREAM_ERROR = "[message_generator()] There was a downstream error."
MSGGEN_PROPAGATING_ERROR = "[message_generator()] Propagating error..."
MSGGEN_EXCEPTED_DOWNSTREAM_ERROR = (
    "[message_generator()] Excepted downstream error (not re-raising):"
)
MSGGEN_GENERATOR_EXITING = "[message_generator()] Exiting generator..."
MSGGEN_GENERATOR_EXITED = "[message_generator()] Exited generator."
MSGGEN_CLOSED_QUEUE = "[message_generator()] Closed queue."



"""Fixtures."""


import os

import pytest_asyncio
from rest_tools.client import ClientCredentialsAuth


def do_skip_auth() -> bool:
    """Return whether to skip all the auth setup."""
    if os.getenv("PYTEST_DO_AUTH_FOR_MQCLIENT", None) == "no":
        return True
    elif os.getenv("PYTEST_DO_AUTH_FOR_MQCLIENT", None) != "yes":
        raise ValueError(
            f"PYTEST_DO_AUTH_FOR_MQCLIENT must be 'yes' or 'no' ({os.getenv('PYTEST_DO_AUTH_FOR_MQCLIENT')})"
        )
    # PYTEST_DO_AUTH_FOR_MQCLIENT is 'yes'
    return False


@pytest_asyncio.fixture
async def auth_token() -> str:
    """Get a valid token from Keycloak test instance."""
    if do_skip_auth():
        return ""

    cc = ClientCredentialsAuth(
        "",
        token_url=os.environ["KEYCLOAK_OIDC_URL"],
        client_id=os.environ["KEYCLOAK_CLIENT_ID"],
        client_secret=os.environ["KEYCLOAK_CLIENT_SECRET"],
    )
    token = cc.make_access_token()
    return token
